import pytest

from gobbli.model.transformer import Transformer


@pytest.mark.parametrize(
    "params,exception",
    [
        # Unknown param
        ({"unknown": None}, ValueError),
        # Bad type (max_seq_length)
        ({"max_seq_length": "100"}, TypeError),
        # OK type (max_seq_length)
        ({"max_seq_length": 100}, None),
        # Bad type (config_overrides)
        ({"config_overrides": 1}, TypeError),
        # OK type (config_overrides)
        ({"config_overrides": {}}, None),
        # Bad type (lr)
        ({"lr": 1}, TypeError),
        # OK type (lr)
        ({"lr": 1e-3}, None),
        # Bad type (adam_eps)
        ({"adam_eps": 1}, TypeError),
        # OK type (adam_eps)
        ({"adam_eps": 1e-5}, None),
        # Bad type (gradient_accumulation_steps)
        ({"gradient_accumulation_steps": 1.0}, TypeError),
        # OK type (gradient_accumulation_steps)
        ({"gradient_accumulation_steps": 2}, None),
        # OK values (all params),
        (
            {
                "max_seq_length": 100,
                "config_overrides": {},
                "lr": 1e-3,
                "adam_eps": 1e-5,
                "gradient_accumulation_steps": 2,
            },
            None,
        ),
    ],
)
def test_init(params, exception):
    if exception is None:
        Transformer(**params)
    else:
        with pytest.raises(exception):
            Transformer(**params)



import inspect
from pathlib import Path

import pytest

from gobbli.interactive.util import read_data_file


@pytest.mark.parametrize(
    "filename,contents,limit,expected",
    [
        # CSV, no text column
        ("test.csv", "a,b", None, ValueError),
        # CSV, no label column
        ("test.csv", "text,b\ntest1,1\ntest2,2", None, (["test1", "test2"], None)),
        # CSV
        (
            "test.csv",
            "text,label\ntest1,1\ntest2,2",
            None,
            (["test1", "test2"], ["1", "2"]),
        ),
        # CSV, limit
        ("test.csv", "text,label\ntest1,1\ntest2,2", 1, (["test1"], ["1"])),
        # TSV, no text column
        ("test.tsv", "a\tb", None, ValueError),
        # TSV, no label column
        ("test.tsv", "text\tb\ntest1\t1\ntest2\t2", None, (["test1", "test2"], None)),
        # TSV
        (
            "test.tsv",
            "text\tlabel\ntest1\t1\ntest2\t2",
            None,
            (["test1", "test2"], ["1", "2"]),
        ),
        # TSV, limit
        ("test.tsv", "text\tlabel\ntest1\t1\ntest2\t2", 1, (["test1"], ["1"])),
        # TXT
        ("test.txt", "test1\ntest2", None, (["test1", "test2"], None)),
        # TXT, limit
        ("test.txt", "test1\ntest2", 1, (["test1"], None)),
    ],
)
def test_read_data_file(tmpdir, filename, contents, limit, expected):
    filepath = Path(tmpdir) / filename
    filepath.write_text(contents)

    if inspect.isclass(expected) and issubclass(expected, Exception):
        with pytest.raises(expected):
            read_data_file(filepath, limit)
    else:
        texts, labels = read_data_file(filepath, limit)
        expected_texts, expected_labels = expected

        assert texts == expected_texts
        assert labels == expected_labels



import tempfile
import warnings
from pathlib import Path

import pandas as pd
import pytest

from gobbli.dataset.newsgroups import NewsgroupsDataset
from gobbli.experiment.classification import (
    ClassificationExperiment,
    ClassificationExperimentResults,
)
from gobbli.inspect.evaluate import DEFAULT_METRICS
from gobbli.model.bert import BERT
from gobbli.model.fasttext import FastText
from gobbli.model.majority import MajorityClassifier
from gobbli.model.mtdnn import MTDNN
from gobbli.model.use import USE
from gobbli.util import dir_to_blob


def test_classification_results_checkpoint(tmpdir):
    # Verify checkpoints can be extracted correctly regardless of format
    tempdir_path = Path(tmpdir)
    checkpoint_path = tempdir_path / "test_checkpoint"
    checkpoint_path.mkdir(parents=True)
    checkpoint_file = checkpoint_path / "checkpoint.txt"

    checkpoint_contents = "test"
    checkpoint_file.write_text(checkpoint_contents)
    checkpoint_bytes = dir_to_blob(checkpoint_path)

    common_args = {
        "training_results": [],
        "labels": [],
        "X": [],
        "y_true": [],
        "y_pred_proba": pd.DataFrame(),
    }
    bytes_results = ClassificationExperimentResults(
        **common_args,
        best_model_checkpoint=checkpoint_bytes,
        best_model_checkpoint_name=checkpoint_file.name,
    )

    path_results = ClassificationExperimentResults(
        **common_args,
        best_model_checkpoint=checkpoint_path,
        best_model_checkpoint_name=checkpoint_file.name,
    )

    # Bytes checkpoint, no base_path (results object creates tempdir)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            "No base_path provided; checkpoint extracting to temporary directory.",
        )
        bytes_checkpoint = bytes_results.get_checkpoint()
    assert bytes_checkpoint.read_text() == checkpoint_contents

    # Bytes checkpoint, base path
    with tempfile.TemporaryDirectory() as test_dir:
        test_dir_path = Path(test_dir) / "test"
        bytes_checkpoint = bytes_results.get_checkpoint(base_path=test_dir_path)
        assert bytes_checkpoint.parent == test_dir_path
        assert bytes_checkpoint.read_text() == checkpoint_contents

    # Path checkpoint, no base path
    path_checkpoint = path_results.get_checkpoint()
    assert path_checkpoint == checkpoint_path / checkpoint_file
    assert path_checkpoint.read_text() == checkpoint_contents

    # Path checkpoint, base path
    with tempfile.TemporaryDirectory() as test_dir:
        test_dir_path = Path(test_dir) / "test"
        path_checkpoint = path_results.get_checkpoint(base_path=test_dir_path)
        assert path_checkpoint.parent == test_dir_path
        assert path_checkpoint.read_text() == checkpoint_contents


@pytest.mark.parametrize(
    "model_cls,valid", [(FastText, True), (BERT, True), (MTDNN, True), (USE, False)]
)
def test_classification_init_validation(model_cls, valid):
    if valid:
        ClassificationExperiment._validate_model_cls(model_cls)
    else:
        with pytest.raises(ValueError):
            ClassificationExperiment._validate_model_cls(model_cls)


@pytest.mark.parametrize(
    "bad_value",
    [
        # Not enough values
        ((0.8, 0.2),),
        # Too many values
        ((0.6, 0.2, 0.1, 0.1),),
        # sum > 1
        ((0.7, 0.2, 0.2),),
        # sum < 1
        ((0.6, 0.2, 0.1),),
    ],
)
def test_classification_run_validation(bad_value):
    with pytest.raises(ValueError):
        ClassificationExperiment._validate_split(bad_value)


@pytest.mark.parametrize(
    "model_cls,dataset_cls,param_grid,limit,ray_local_mode",
    [
        # Can't use the TrivialDataset here because it's too small for the standard
        # train/valid/test split
        # Trivial model, no ray
        (MajorityClassifier, NewsgroupsDataset, {}, 1000, True),
        # "Real" model/dataset, use ray cluster
        (
            FastText,
            NewsgroupsDataset,
            {"lr": [0.1, 0.01], "autotune_duration": [5]},
            200,
            False,
        ),
        # "Real" model/dataset with more complicated checkpoint structure, use ray cluster
        # Use smaller limit since this model takes a while to train
        (BERT, NewsgroupsDataset, {}, 50, False),
    ],
)
def test_classification_run(
    request, model_cls, dataset_cls, param_grid, limit, ray_local_mode, gobbli_dir
):
    if model_cls == BERT:
        pytest.skip(
            "BERT model takes up too much disk space; this test is currently disabled"
        )

    dataset = dataset_cls.load()

    exp = ClassificationExperiment(
        model_cls=model_cls,
        dataset=dataset,
        param_grid=param_grid,
        task_num_cpus=1,
        task_num_gpus=0,
        worker_gobbli_dir=gobbli_dir,
        worker_log_level=request.config.getoption("worker_log_level"),
        limit=limit,
        ignore_ray_initialized_error=True,
        ray_kwargs={"local_mode": ray_local_mode},
    )

    results = exp.run()

    if not model_cls == MajorityClassifier:
        assert results.best_model_checkpoint is not None
        assert results.best_model_checkpoint_name is not None

    metrics = results.metrics()
    assert len(metrics) == len(DEFAULT_METRICS)
    for metric, value in metrics.items():
        assert isinstance(value, float)

    metrics_report = results.metrics_report()
    assert len(metrics_report) > 0

    k = 5
    errors = results.errors(k=k)
    for label, (false_positives, false_negatives) in errors.items():
        assert len(false_positives) <= k
        assert len(false_negatives) <= k

    errors_report = results.errors_report(k=k)
    assert len(errors_report) > 0

    # Verify the plot runs without errors
    results.plot()



import distutils.file_util
import enum
import gzip
import io
import json
import logging
import os
import random
import shutil
import tarfile
import tempfile
import uuid
import warnings
import zipfile
from distutils.util import strtobool
from pathlib import Path
from typing import Any, Container, Dict, Iterator, List, Optional, TypeVar

import humanize
import pandas as pd
import pkg_resources
import requests

LOGGER = logging.getLogger(__name__)


def default_gobbli_dir() -> Path:
    """
    Returns:
      The default directory to be used to store gobbli data if there's no user-specified
      default.
    """
    return Path.home() / ".gobbli"


def gobbli_dir() -> Path:
    """
    Returns:
      The directory used to store gobbli data.  Can be overridden using the `GOBBLI_DIR`
      environment variable.
    """
    gobbli_dir = Path(os.getenv("GOBBLI_DIR", str(default_gobbli_dir())))
    gobbli_dir.mkdir(parents=True, exist_ok=True)
    return gobbli_dir


def model_dir() -> Path:
    """
    Returns:
      The subdirectory storing all model data and task run input/output.
    """
    return gobbli_dir() / "model"


def gobbli_version() -> str:
    """
    Returns:
      The version of gobbli installed.
    """
    return pkg_resources.get_distribution("gobbli").version


def disk_usage() -> int:
    """
    Returns:
      Disk usage (in bytes) of all gobbli data in the current gobbli directory.
    """
    return sum(f.stat().st_size for f in gobbli_dir().glob("**/*") if f.is_file())


def human_disk_usage() -> str:
    """
    Returns:
      Human-readable string representing disk usage of all gobbli data in the current gobbli
      directory.
    """
    return humanize.naturalsize(disk_usage())


def cleanup(force: bool = False, full: bool = False):
    """
    Cleans up the gobbli directory, removing files and directories.  By default, removes
    only task input/output.  If the ``full`` argument is True, removes all data,
    including downloaded model weights and datasets.

    Args:
      force: If True, don't prompt for user confirmation before cleaning up.
      full: If True, remove all data (including downloaded model weights and datasets).
    """
    if not force:
        msg = "Cleanup will remove all task input/output, including trained models.  Are you sure? [Y/n]"
        confirmation = bool(strtobool(input(msg)))
        if not confirmation:
            return

        if full:
            full_msg = "Full cleanup will additionally remove all downloaded model weights; they'll need to be redownloaded to be used again.  Are you sure? [Y/n]"
            confirmation = bool(strtobool(input(full_msg)))
            if not confirmation:
                return

    if full:
        shutil.rmtree(gobbli_dir())
    else:
        for model_dir in (gobbli_dir() / "model").iterdir():
            for instance_dir in model_dir.iterdir():
                # Special directories containing cached weights or models
                # which are reusable -- don't delete them here
                if instance_dir.name not in ("cache", "weights"):
                    shutil.rmtree(instance_dir)


def pred_prob_to_pred_label(y_pred_proba: pd.DataFrame) -> List[str]:
    """
    Convert a dataframe of predicted probabilities (shape (n_samples, n_classes)) to
    a list of predicted classes.
    """
    return y_pred_proba.idxmax(axis=1).tolist()


def truncate_text(text: str, length: int) -> str:
    """
    Truncate the text to the given length.  Append an ellipsis to make it clear the
    text was truncated.

    Args:
      text: The text to truncate.
      length: Maximum length of the truncated text.

    Returns:
      The truncated text.
    """
    truncated = text[:length]
    if len(truncated) < len(text):
        truncated += "..."
    return truncated


def assert_type(name: str, val: Any, cls: Any):
    """
    Raise an informative error if the passed value isn't of the given type.

    Args:
      name: User-friendly name of the value to be printed in an exception if raised.
      val: The value to check type of.
      cls: The type or tuple of types to compare the value's type against.
    """
    if not isinstance(val, cls):
        raise TypeError(f"{name} must be of type(s) '{cls}'; got '{type(val)}'")


def assert_in(name: str, val: Any, container: Container):
    """
    Raise an informative error if the passed value isn't in the given container.

    Args:
      name: User-friendly name of the value to be printed in an exception if raised.
      val: The value to check for membership in the container.
      container: The container that should contain the value.
    """
    if val not in container:
        raise ValueError(f"invalid {name} '{val}'; valid values are {container}")


def generate_uuid() -> str:
    """
    Generate a universally unique ID to be used for randomly naming directories,
    models, tasks, etc.
    """
    return uuid.uuid4().hex


def download_dir() -> Path:
    """
    Returns:
      The directory used to store gobbli downloaded files.
    """
    download_dir = gobbli_dir() / "download"
    download_dir.mkdir(parents=True, exist_ok=True)
    return download_dir


def escape_line_delimited_text(text: str) -> str:
    """
    Convert a single text possibly containing newlines and other troublesome whitespace
    into a string suitable for writing and reading to a file where newlines will
    divide up the texts.

    Args:
      text: The text to convert.

    Returns:
      The text with newlines and whitespace taken care of.
    """
    return text.replace("\n", " ").strip()


def escape_line_delimited_texts(texts: List[str]) -> str:
    """
    Convert a list of texts possibly containing newlines and other troublesome whitespace
    into a string suitable for writing and reading to a file where newlines
    will divide up the texts.

    Args:
      texts: The list of texts to convert.

    Returns:
      The newline-delimited string.
    """
    clean_texts = [escape_line_delimited_text(text) for text in texts]
    return "\n".join(clean_texts)


def is_dir_empty(dir_path: Path) -> bool:
    """
    Determine whether a given directory is empty.  Assumes the directory exists.

    Args:
      dir_path: The directory to check for emptiness.

    Returns:
      :obj:`True` if the directory is empty, otherwise :obj:`False`.
    """
    empty = True
    for child in dir_path.iterdir():
        empty = False
        break
    return empty


def write_metadata(metadata: Dict[str, Any], file_path: Path):
    """
    Write some JSON-formatted metadata to the given file path,
    formatted appropriately for human consumption.

    Args:
      metadata: The valid JSON metadata to write.  Nesting is allowed, but
        ensure all types in the structure are JSON-serializable.
      file_path: The path to write the metadata to.
    """
    with file_path.open("w") as f:
        json.dump(metadata, f, indent=4)


def read_metadata(file_path: Path) -> Dict[str, Any]:
    """
    Read JSON-formatted metadata from the given file path.

    Args:
      file_path: The path to read JSON metadata from.

    Returns:
      The JSON object read from the file.
    """
    with file_path.open("r") as f:
        return json.load(f)


def format_duration(seconds: float) -> str:
    """
    Nicely format a given duration in seconds.

    Args:
      seconds: The duration to format, in seconds.

    Returns:
      The duration formatted as a string with unit of measurement appended.
    """
    return f"{seconds:.2f} sec"


def copy_file(src_path: Path, dest_path: Path) -> bool:
    """
    Copy a file from the source to destination.  Be smart -- only copy if
    the source is newer than the destination.

    Args:
      src_path: The path to the source file.
      dest_path: The path to the destination where the source should be copied.

    Returns:
      True if the file was copied, otherwise False if the destination was already
      up to date.
    """
    _, copied = distutils.file_util.copy_file(
        str(src_path), str(dest_path), update=True
    )
    return copied == 1


def download_file(url: str, filename: Optional[str] = None) -> Path:
    """
    Save a file in the gobbli download directory if it doesn't already exist there.
    Stream the download to avoid running out of memory.

    Args:
      url: URL for the file.
      filename: If passed, use this as the filename instead of the best-effort one
        determined from the URL.

    Returns:
      The path to the downloaded file.
    """
    if filename is None:
        # Kind of hacky... pull out the last path component as the filename and strip
        # a trailing query, if any
        local_filename = url.split("/")[-1]

        try:
            query_start_ndx = local_filename.index("?")
        except ValueError:
            query_start_ndx = -1

        if query_start_ndx != -1:
            local_filename = local_filename[:query_start_ndx]
    else:
        local_filename = filename

    local_filepath = download_dir() / local_filename

    if local_filepath.exists():
        LOGGER.debug(f"Download for URL '{url}' already exists at '{local_filepath}'")
        return local_filepath

    LOGGER.debug(f"Downloading URL '{url}' to '{local_filepath}'")
    try:
        with requests.get(url, stream=True) as r:
            with open(local_filepath, "wb") as f:
                shutil.copyfileobj(r.raw, f)
    except Exception as e:
        # Don't leave the file in a partially downloaded state
        try:
            local_filepath.unlink()
            LOGGER.debug(f"Removed partially downloaded file at '{local_filepath}'")
        except Exception:
            warnings.warn(
                "Failed to remove partially downloaded file at '{local_filepath}'."
            )
        raise e

    return local_filepath


def _extract_tar_junk_path(tarfile_obj: tarfile.TarFile, archive_extract_dir: Path):
    """
    Extract a tarfile while flattening any directory hierarchy
    in the archive.
    """
    for member in tarfile_obj.getmembers():
        if member.isdir():
            # Skip directories
            continue
        # Remove the directory hierarchy from the file
        member.name = Path(member.name).name
        output_file = archive_extract_dir / member.name
        LOGGER.debug(f"Extracting member '{member.name}' to '{output_file}'")
        tarfile_obj.extract(member, path=archive_extract_dir)


def _extract_zip_junk_path(zipfile_obj: zipfile.ZipFile, archive_extract_dir: Path):
    """
    Extract a zip file while flattening any directory hierarchy in the archive.
    """
    for member in zipfile_obj.infolist():
        if member.is_dir():
            # Skip directories
            continue
        member_name = Path(member.filename).name
        output_file = archive_extract_dir / member_name
        LOGGER.debug(f"Extracting member '{member_name}' to '{output_file}'")
        with zipfile_obj.open(member, "r") as f:
            with output_file.open("wb") as f_out:
                shutil.copyfileobj(f, f_out)


_SUPPORTED_ARCHIVE_EXTENSIONS = (".gz", ".zip")


def is_archive(filepath: Path) -> bool:
    """
    Args:
      filepath: Path to a file.

    Returns:
      Whether the file is an archive supported by :func:`extract_archive`.
    """
    for ext in _SUPPORTED_ARCHIVE_EXTENSIONS:
        if filepath.name.endswith(ext):
            return True
    return False


def extract_archive(
    archive_path: Path, archive_extract_dir: Path, junk_paths: bool = False
):
    """
    Extract an archive to the given directory.

    Args:
      archive_path: Path to the archive file.
      archive_extract_dir: Extract the archive to this directory.
      junk_paths: If True, disregard the archive's internal directory hierarchy
        and extract all files directly to the output directory.

    Returns:
      Path to the directory containing the extracted file contents.
    """
    LOGGER.debug(f"Extracting archive '{archive_path}'")
    archive_extract_dir.mkdir(exist_ok=True, parents=True)

    if archive_path.name.endswith(".tar.gz"):
        with tarfile.open(archive_path, "r:gz") as archive_tar:
            if junk_paths:
                _extract_tar_junk_path(archive_tar, archive_extract_dir)
            else:
                LOGGER.debug(f"Extracting all members to '{archive_extract_dir}'")
                archive_tar.extractall(archive_extract_dir)

    elif archive_path.name.endswith(".gz"):
        LOGGER.debug(f"Extracting gzipped file to '{archive_extract_dir}'")
        with gzip.open(archive_path, "rb") as archive_gz:
            # Strip the trailing '.gz' and use the original filename as the new filename
            with open(archive_extract_dir / archive_path.name[:-3], "wb") as f:
                shutil.copyfileobj(archive_gz, f)

    elif archive_path.name.endswith(".zip"):
        with zipfile.ZipFile(archive_path, "r") as archive_zip:
            if junk_paths:
                _extract_zip_junk_path(archive_zip, archive_extract_dir)
            else:
                LOGGER.debug(f"Extracting all members to '{archive_extract_dir}'")
                archive_zip.extractall(archive_extract_dir)
    else:
        raise ValueError(f"Unsupported archive file: {archive_path}")

    return archive_extract_dir


def download_archive(
    archive_url: str,
    archive_extract_dir: Path,
    junk_paths: bool = False,
    filename: Optional[str] = None,
) -> Path:
    """
    Save an archive in the given directory and extract its contents.  Automatically
    retry the download once if the file comes back corrupted (in case it's left over
    from a partial download that was cancelled before).

    Args:
      archive_url: URL for the archive.
      archive_extract_dir: Download the archive and extract it to this directory.
      junk_paths: If True, disregard the archive's internal directory hierarchy
        and extract all files directly to the output directory.
      filename: If given, store the downloaded file under this name instead of
        one automatically inferred from the URL.

    Returns:
      Path to the directory containing the extracted file contents.
    """
    download_path = download_file(archive_url, filename=filename)
    try:
        return extract_archive(
            download_path, archive_extract_dir, junk_paths=junk_paths
        )
    except (zipfile.BadZipFile, tarfile.ReadError, OSError, EOFError):
        LOGGER.warning(
            f"Downloaded archive at '{download_path}' is corrupted.  Retrying..."
        )
        download_path.unlink()
        download_path = download_file(archive_url, filename=filename)
        return extract_archive(
            download_path, archive_extract_dir, junk_paths=junk_paths
        )


def dir_to_blob(dir_path: Path) -> bytes:
    """
    Archive a directory and save it as a blob in-memory.
    Useful for storing a directory's contents in an in-memory object store.
    Use compression to reduce file size.  Extract with :func:`blob_to_dir`.

    Args:
      dir_path: Path to the directory to be archived.

    Returns:
      The compressed directory as a binary buffer.
    """
    blob = io.BytesIO()
    with tarfile.open(fileobj=blob, mode="w:gz") as archive:
        # Set arcname=. to ensure the files will be extracted properly
        # using relative paths
        archive.add(str(dir_path), arcname=".", recursive=True)

    # Seek the beginning of the buffer so we read the whole thing
    blob.seek(0)
    return blob.getvalue()


def blob_to_dir(blob: bytes, dir_path: Path):
    """
    Extract the given blob (assumed to be a compressed directory
    created by :func:`dir_to_blob`) to the given directory.

    Args:
      blob: The compressed directory as a binary buffer.
      dir_path: Path to extract the directory to.
    """
    with tarfile.open(fileobj=io.BytesIO(blob), mode="r:gz") as archive:
        archive.extractall(dir_path)


T1 = TypeVar("T1")
T2 = TypeVar("T2")


def shuffle_together(l1: List[T1], l2: List[T2], seed: Optional[int] = None):
    """
    Shuffle two lists together, so their order is random but the individual
    elements still correspond.  Ex. shuffle a list of texts and a list of labels
    so the labels still correspond correctly to the texts.  The lists are shuffled in-place.

    The lists must be the same length for this to make sense.

    Args:
      l1: The first list to be shuffled.
      l2: The second list to be shuffled.
      seed: Seed for the random number generator, if any
    """
    if not len(l1) == len(l2):
        raise ValueError("Lists have unequal length.")
    if len(l1) == 0:
        return

    zipped = list(zip(l1, l2))
    if seed is not None:
        random.seed(seed)
    random.shuffle(zipped)
    l1[:], l2[:] = zip(*zipped)


def _train_sentencepiece(spm: Any, texts: List[str], model_path: Path, vocab_size: int):
    """
    Train a Sentencepiece model on the given data and save it to the given
    location.

    Args:
      spm: Imported sentencepiece module
      texts: Data to train on.
      model_path: Path to write the trained model to.
      vocab_size: Size of the vocabulary for the model to train.
    """
    with tempfile.NamedTemporaryFile(mode="w") as f:
        f.write("\n".join(texts))
        f.flush()

        # log levels:
        # https://github.com/google/sentencepiece/blob/d4dd947fe71c4fa4ee24ad8297beee32887d8828/src/common.h#L132
        # Default is waaay too chatty
        spm.SentencePieceTrainer.train(
            f"--input={f.name} --model_prefix={model_path} --vocab_size={vocab_size} --minloglevel=2"
        )


@enum.unique
class TokenizeMethod(enum.Enum):
    """
    Enum describing the different canned tokenization methods gobbli supports.
    Processes requiring tokenization should generally allow a user to pass in
    a custom tokenization function if their needs aren't met by one of these.

    Attributes:
      SPLIT: Naive tokenization based on whitespace.  Probably only useful for testing.
        Tokens will be lowercased.
      SPACY: Simple tokenization using spaCy's English language model.
        Tokens will be lowercased, and non-alphabetic tokens will be filtered out.
      SENTENCEPIECE: `SentencePiece <https://github.com/google/sentencepiece>`__-based tokenization.
    """

    SPLIT = "split"
    SPACY = "spacy"
    SENTENCEPIECE = "sentencepiece"


def tokenize(
    method: TokenizeMethod,
    texts: List[str],
    model_path: Optional[Path] = None,
    vocab_size: int = 2000,
) -> List[List[str]]:
    """
    Tokenize a list of texts using a predefined method.

    Args:
      texts: Texts to tokenize.
      method: The type of tokenization to apply.
      model_path: Path to save a trained model to.  Required if the tokenization method
        requires training a model; otherwise ignored.  If the model doesn't exist, it will
        be trained; if it does, the trained model will be reused.
      vocab_size: Number of terms in the vocabulary for tokenization methods with a fixed
        vocabulary size. You may need to lower this if you get tokenization errors or
        raise it if your texts have a very diverse vocabulary.
    Returns:
      List of tokenized texts.
    """
    if method == TokenizeMethod.SPLIT:
        return [[tok.lower() for tok in text.split()] for text in texts]
    elif method == TokenizeMethod.SPACY:
        try:
            from spacy.lang.en import English
        except ImportError:
            raise ImportError(
                "spaCy tokenization method requires spaCy and an english language "
                "model to be installed."
            )
        nlp = English()
        tokenizer = nlp.Defaults.create_tokenizer(nlp)
        processed_texts = []
        for doc in tokenizer.pipe(texts):
            processed_texts.append([tok.lower_ for tok in doc if tok.is_alpha])
        return processed_texts
    elif method == TokenizeMethod.SENTENCEPIECE:
        try:
            import sentencepiece as spm
        except ImportError:
            raise ImportError(
                "SentencePiece tokenization requires the sentencepiece module "
                "to be installed."
            )
        # Train only if the model file path doesn't already exist
        # If we weren't given a path to save to, just make a temp file which will
        # be discarded after the run
        sp = spm.SentencePieceProcessor()
        if model_path is None:
            with tempfile.TemporaryDirectory() as temp_model_dir:
                temp_model_path = Path(temp_model_dir) / "temp"
                _train_sentencepiece(spm, texts, temp_model_path, vocab_size)
                # Have to add the extension here for sentencepiece to find the model
                sp.load(f"{temp_model_path}.model")
                return [sp.encode_as_pieces(text) for text in texts]
        else:
            model_file_path = Path(f"{model_path}.model")

            if not model_file_path.exists():
                _train_sentencepiece(spm, texts, model_path, vocab_size)
            sp.load(str(model_file_path))
            return [sp.encode_as_pieces(text) for text in texts]
    else:
        raise ValueError(f"Unsupported tokenization method '{method}'.")


def detokenize(
    method: TokenizeMethod,
    all_tokens: Iterator[List[str]],
    model_path: Optional[Path] = None,
) -> List[str]:
    """
    Detokenize a nested list of tokens into a list of strings, assuming
    they were created using the given predefined method.

    Args:
      method: The type of tokenization to reverse.
      tokens: The nested list of tokens to detokenize.
      model_path: Path to load a trained model from.  Required if the tokenization method
        requires training a model; otherwise ignored.
    Returns:
      List of texts.
    """
    if method in (TokenizeMethod.SPLIT, TokenizeMethod.SPACY):
        # TODO can this be better?
        return [" ".join(tokens) for tokens in all_tokens]
    elif method == TokenizeMethod.SENTENCEPIECE:
        try:
            import sentencepiece as spm
        except ImportError:
            raise ImportError(
                "SentencePiece detokenization requires the sentencepiece module "
                "to be installed."
            )
        sp = spm.SentencePieceProcessor()
        sp.load(f"{model_path}.model")
        return [sp.decode_pieces(tokens) for tokens in all_tokens]
    else:
        raise ValueError(f"Unsupported tokenization method '{method}'.")



from .model import (
    SKLearnClassifier,
    TfidfEmbedder,
    make_cv_tfidf_logistic_regression,
    make_default_tfidf_logistic_regression,
    persist_estimator,
)

__all__ = [
    "SKLearnClassifier",
    "TfidfEmbedder",
    "persist_estimator",
    "make_cv_tfidf_logistic_regression",
    "make_default_tfidf_logistic_regression",
]



import logging
import math
import shutil
import tempfile
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union, cast

import altair as alt
import pandas as pd
import ray
from sklearn.model_selection import ParameterGrid, train_test_split

import gobbli.io
from gobbli.experiment.base import (
    BaseExperiment,
    get_worker_ip,
    init_gpu_config,
    init_worker_env,
    is_ray_local_mode,
)
from gobbli.inspect.evaluate import ClassificationError, ClassificationEvaluation
from gobbli.model.mixin import PredictMixin, TrainMixin
from gobbli.util import blob_to_dir, dir_to_blob


@dataclass
class ClassificationExperimentResults:
    """
    Results from a classification experiment.  An experiment entails training a set of models
    based on a grid of parameters, retraining on the full train/validation dataset with the
    best set of parameters, and evaluating predictions on the test set.

    Args:
      training_results: A list of dictionaries containing information about each training run,
        one for each unique combination of hyperparameters in :paramref:`BaseExperiment.params.param_grid`.
      labels: The set of unique labels in the dataset.
      X: The list of texts to classify.
      y_true: The true labels for the test set, as passed by the user.
      y_pred_proba: A dataframe containing a row for each observation in the test set and a
        column for each label in the training data.  Cells are predicted probabilities.
      best_model_checkpoint: If results came from another process on the master node, this is
        the directory containing the checkpoint.  If the results came from a worker node, this
        is a bytes object containing the compressed model weights.
      best_model_checkpoint_name: Path to the best checkpoint within the directory or
        or compressed blob.
      metric_funcs: Overrides for the default set of metric functions used to evaluate
        the classifier's performance.
    """

    training_results: List[Dict[str, Any]]
    labels: List[str]
    X: List[str]
    y_true: List[str]
    y_pred_proba: pd.DataFrame
    best_model_checkpoint: Union[bytes, Path]
    best_model_checkpoint_name: str
    metric_funcs: Optional[Dict[str, Callable[[Sequence, Sequence], float]]] = None

    def __post_init__(self):
        self.evaluation = ClassificationEvaluation(
            labels=self.labels,
            X=self.X,
            y_true=self.y_true,
            y_pred_proba=self.y_pred_proba,
            metric_funcs=self.metric_funcs,
        )

    def get_checkpoint(self, base_path: Optional[Path] = None) -> Path:
        """
        Return a filesystem path to our checkpoint, which can be used to initialize
        future models from the same state. If a base_path is provided, copy/extract
        the checkpoint under that path.

        NOTE: If no base_path is provided and the checkpoint comes from a remote
        worker, the checkpoint will be extracted to a temporary directory, and a
        warning will be emitted.  gobbli will make no effort to ensure the temporary
        directory is cleaned up after creation.

        Args:
          base_path: Optional directory to extract/copy the checkpoint to. If not provided,
            the original path will be returned if the checkpoint already existed on the
            current machine's filesystem.  If the checkpoint is a bytes object, a temporary
            directory will be created.  The directory must not already exist.

        Returns:
          The path to the extracted checkpoint.
        """
        if isinstance(self.best_model_checkpoint, bytes):
            if base_path is None:
                warnings.warn(
                    "No base_path provided; checkpoint extracting to temporary "
                    "directory."
                )
                base_path = Path(tempfile.mkdtemp())

            blob_to_dir(self.best_model_checkpoint, base_path)
            return base_path / self.best_model_checkpoint_name

        elif isinstance(self.best_model_checkpoint, Path):
            if base_path is None:
                base_path = self.best_model_checkpoint
            else:
                # Copy the checkpoint to the user-provided base path
                shutil.copytree(self.best_model_checkpoint, base_path)
            return base_path / self.best_model_checkpoint_name
        else:
            raise TypeError(
                f"unsupported checkpoint type: '{type(self.best_model_checkpoint)}'"
            )

    def metrics(self, *args, **kwargs) -> Dict[str, float]:
        """
        See :meth:`ClassificationEvaluation.metrics`.
        """
        return self.evaluation.metrics(*args, **kwargs)

    def metrics_report(self, *args, **kwargs) -> str:
        """
        See :meth:`ClassificationEvaluation.metrics_report`.
        """
        return self.evaluation.metrics_report(*args, **kwargs)

    def plot(self, *args, **kwargs) -> alt.Chart:
        """
        See :meth:`ClassificationEvaluation.plot`.
        """
        return self.evaluation.plot(*args, **kwargs)

    def errors(
        self, *args, **kwargs
    ) -> Dict[str, Tuple[List[ClassificationError], List[ClassificationError]]]:
        """
        See :meth:`ClassificationEvaluation.errors`.
        """
        return self.evaluation.errors(*args, **kwargs)

    def errors_report(self, *args, **kwargs) -> str:
        """
        See :meth:`ClassificationEvaluation.errors_report`.
        """
        return self.evaluation.errors_report(*args, **kwargs)


@dataclass
class RemoteTrainResult:
    """
    Results from a training process on a (possibly remote) worker.

    Args:
      metadata: Metadata from the training output.
      labels: List of labels identified in the data.
      checkpoint_name: Name of the checkpoint under the checkpoint directory.
      checkpoint_id: ray ObjectID for the checkpoint directory (path or bytes)
      model_params: Parameters used to initialize the model.
      ip_address: IP address of the node that ran training.
    """

    metadata: Dict[str, Any]
    labels: List[str]
    checkpoint_name: str
    checkpoint_id: ray.ObjectID
    model_params: Dict[str, Any]
    ip_address: str


class ClassificationExperiment(BaseExperiment):
    """
    Run a classification experiment.  This entails training a model to
    make predictions given some input.

    The experiment will, for each combination of model hyperparameters,
    train the model on a training set and evaluate it on a validation set.
    The best combination of hyperparameters will be retrained on the combined
    training/validation sets and evaluated on the test set. After completion,
    the experiment will return :class:`ClassificationExperimentResults`, which will
    allow the user to examine the results in various ways.
    """

    _DEFAULT_TRAIN_VALID_TEST_SPLIT = (0.7, 0.1, 0.2)
    _DEFAULT_TRAIN_VALID_SPLIT = (0.8, 0.2)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        ClassificationExperiment._validate_model_cls(self.model_cls)

    @staticmethod
    def _validate_model_cls(model_cls: Any):
        if not (
            issubclass(model_cls, TrainMixin) and issubclass(model_cls, PredictMixin)
        ):
            raise ValueError(
                f"Model of class {model_cls.__name__} can't be used for a"
                " classification experiment because it doesn't support both"
                " training and prediction."
            )

    @staticmethod
    def _validate_split(
        train_valid_test_split: Union[Tuple[float, float, float], Tuple[float, float]],
        expected_len=3,
    ):
        if (
            not len(train_valid_test_split) == expected_len
            or sum(train_valid_test_split) != 1
        ):
            raise ValueError(
                f"train_valid_test_split must be length {expected_len} and sum to 1"
            )

    def run(
        self,
        dataset_split: Optional[
            Union[Tuple[float, float], Tuple[float, float, float]]
        ] = None,
        seed: int = 1,
        train_batch_size: int = 32,
        valid_batch_size: int = 32,
        test_batch_size: int = 32,
        num_train_epochs: int = 5,
    ) -> ClassificationExperimentResults:
        """
        Run the experiment.

        Args:
          dataset_split: A tuple describing the proportion of the dataset
            to be added to the train/validation/test splits.  If the experiment uses an explicit
            test set (passes :paramref:`BaseExperiment.params.test_dataset`), this should be a
            2-tuple describing the train/validation split.  Otherwise, it should be a 3-tuple
            describing the train/validation/test split. The tuple must sum to 1.
          seed: Random seed to be used for dataset splitting for reproducibility.
          train_batch_size: Number of observations per batch on the training dataset.
          valid_batch_size: Number of observations per batch on the validation dataset.
          test_batch_size: Number of observations per batch on the test dataset.
          num_train_epochs: Number of epochs to use for training.

        Returns:
          The results of the experiment.
        """
        _dataset_split = dataset_split

        # If the user didn't pass an explicit test set, create one
        # using a split
        if self.X_test is None:
            if _dataset_split is None:
                _dataset_split = (
                    ClassificationExperiment._DEFAULT_TRAIN_VALID_TEST_SPLIT
                )

            ClassificationExperiment._validate_split(_dataset_split, expected_len=3)

            # cast needed to satisfy mypy
            train_prop, valid_prop, test_prop = cast(
                Tuple[float, float, float], _dataset_split
            )
            train_valid_prop = train_prop + valid_prop

            X_train_valid, X_test, y_train_valid, y_test = train_test_split(
                self.X, self.y, train_size=train_valid_prop, test_size=test_prop
            )

        else:
            if _dataset_split is None:
                _dataset_split = ClassificationExperiment._DEFAULT_TRAIN_VALID_SPLIT

            ClassificationExperiment._validate_split(_dataset_split, expected_len=2)

            # cast needed to satisfy mypy
            train_prop, valid_prop = cast(Tuple[float, float], _dataset_split)
            train_valid_prop = 1

            X_train_valid, y_train_valid = self.X, self.y
            X_test, y_test = self.X_test, self.y_test

        X_train, X_valid, y_train, y_valid = train_test_split(
            X_train_valid,
            y_train_valid,
            # Round to prevent floating point imprecision errors
            train_size=round(train_prop / train_valid_prop, 4),
            test_size=round(valid_prop / train_valid_prop, 4),
        )

        if self.param_grid is not None:
            for param, values in self.param_grid.items():
                if isinstance(values, str):
                    raise TypeError(
                        f"String detected in parameter grid values for parameter '{param}'. "
                        "This will be treated as a list of character parameter values, "
                        "which probably isn't what you want.  If you're really sure, "
                        "convert the string to a list of characters and try again."
                    )
        grid = ParameterGrid(self.param_grid)
        if len(grid) == 0:
            raise ValueError("empty parameter grid")

        # Transfer datasets to the Ray distributed object store
        # if not running in local mode
        # In local mode, this causes problems: https://github.com/ray-project/ray/issues/5379
        if is_ray_local_mode():
            dataset_ids = [X_train, y_train, X_valid, y_valid]
        else:
            dataset_ids = [ray.put(d) for d in (X_train, y_train, X_valid, y_valid)]

        # Return the checkpoint blob separately from the train result so it doesn't
        # have to be copied to the object store again when used by the predict function
        @ray.remote(num_cpus=self.task_num_cpus, num_gpus=self.task_num_gpus)
        def train(
            X_train: Any,
            y_train: Any,
            X_valid: Any,
            y_valid: Any,
            train_batch_size: int,
            valid_batch_size: int,
            num_train_epochs: int,
            model_cls: Any,
            model_params: Dict[str, Any],
            master_ip: str,
            gobbli_dir: Optional[Path] = None,
            log_level: Union[int, str] = logging.WARNING,
            distributed: bool = False,
        ) -> RemoteTrainResult:

            logger = init_worker_env(gobbli_dir=gobbli_dir, log_level=log_level)
            use_gpu, nvidia_visible_devices = init_gpu_config()

            worker_ip = get_worker_ip()
            if not distributed and worker_ip != master_ip:
                raise RuntimeError(
                    "Experiments must be started with distributed = True to run "
                    "tasks on remote workers."
                )

            clf = model_cls(
                **model_params,
                use_gpu=use_gpu,
                nvidia_visible_devices=nvidia_visible_devices,
                logger=logger,
            )

            clf.build()

            train_input = gobbli.io.TrainInput(
                X_train=X_train,
                y_train=y_train,
                X_valid=X_valid,
                y_valid=y_valid,
                train_batch_size=train_batch_size,
                valid_batch_size=valid_batch_size,
                num_train_epochs=num_train_epochs,
            )
            train_output = clf.train(train_input)
            checkpoint = train_output.checkpoint
            checkpoint_name = getattr(checkpoint, "name", None)

            if distributed:
                # Copy weights into the object store, since we don't share a filesystem
                # with the master node
                checkpoint = (
                    dir_to_blob(checkpoint.parent) if checkpoint is not None else None
                )

            if not is_ray_local_mode():
                checkpoint = ray.put(checkpoint)

            return RemoteTrainResult(
                metadata=train_output.metadata(),
                labels=train_output.labels,
                checkpoint_name=checkpoint_name,
                checkpoint_id=checkpoint,
                model_params=model_params,
                ip_address=worker_ip,
            )

        @ray.remote(num_cpus=self.task_num_cpus, num_gpus=self.task_num_gpus)
        def predict(
            X_test: List[str],
            test_batch_size: int,
            model_cls: Any,
            model_params: Dict[str, Any],
            labels: List[str],
            checkpoint: Union[bytes, Path],
            checkpoint_name: Optional[str],
            master_ip: str,
            gobbli_dir: Optional[Path] = None,
            log_level: Union[int, str] = logging.WARNING,
            distributed: bool = False,
        ) -> pd.DataFrame:

            logger = init_worker_env(gobbli_dir=gobbli_dir, log_level=log_level)
            use_gpu, nvidia_visible_devices = init_gpu_config()

            worker_ip = get_worker_ip()
            if not distributed and worker_ip != master_ip:
                raise RuntimeError(
                    "Experiments must be started with distributed = True to run "
                    "tasks on remote workers."
                )

            clf = model_cls(
                **model_params,
                use_gpu=use_gpu,
                nvidia_visible_devices=nvidia_visible_devices,
                logger=logger,
            )

            # This step isn't necessary in all cases if the build step just downloads
            # pretrained weights we weren't going to use anyway, but sometimes it's needed
            # Ex. for BERT to download vocabulary files and config
            clf.build()

            # Use the current working directory (CWD) as the base for the tempdir, under the
            # assumption that the CWD is included in any bind mounts/volumes the user may have
            # created if they're running this in a Docker container
            # If it's not part of a host mount, the files won't be mounted properly in the container
            with tempfile.TemporaryDirectory(dir=".") as tempdir:
                tempdir_path = Path(tempdir)

                checkpoint_path = None  # type: Optional[Path]
                if isinstance(checkpoint, bytes):
                    if checkpoint_name is not None:
                        blob_to_dir(checkpoint, tempdir_path)
                        checkpoint_path = tempdir_path / checkpoint_name
                elif isinstance(checkpoint, Path):
                    checkpoint_path = checkpoint
                elif checkpoint is None:
                    pass
                else:
                    raise TypeError(f"invalid checkpoint type: '{type(checkpoint)}'")

                predict_input = gobbli.io.PredictInput(
                    X=X_test,
                    labels=labels,
                    checkpoint=checkpoint_path,
                    predict_batch_size=test_batch_size,
                )
                predict_output = clf.predict(predict_input)

            return predict_output.y_pred_proba

        # Record the IP address of the master node so workers can detect
        # whether they're remote and not running in distributed mode, at which
        # point they should raise an error
        master_ip = get_worker_ip()

        # Run training in parallel using the Ray cluster
        raw_results = ray.get(
            [
                train.remote(
                    *dataset_ids,
                    train_batch_size,
                    valid_batch_size,
                    num_train_epochs,
                    self.model_cls,
                    params,
                    master_ip,
                    self.worker_gobbli_dir,
                    self.worker_log_level,
                    self.distributed,
                )
                for params in grid
            ]
        )

        training_results = []  # type: List[Dict[str, Any]]
        best_valid_loss = math.inf
        best_result = None  # type: Optional[RemoteTrainResult]
        best_checkpoint_id = None  # type: Optional[ray.ObjectID]

        for train_results in raw_results:
            result = {
                **train_results.metadata,
                "node_ip_address": train_results.ip_address,
                "model_params": train_results.model_params,
            }
            if result["valid_loss"] < best_valid_loss:
                best_result = train_results
                best_checkpoint_id = train_results.checkpoint_id
                best_valid_loss = result["valid_loss"]

            training_results.append(result)

        if best_result is None:
            raise ValueError(
                "failed to find parameter combination with finite validation loss"
            )

        # Evaluate the best model on the test set
        if is_ray_local_mode():
            X_test_id = X_test
        else:
            X_test_id = ray.put(X_test)
        y_pred_proba = ray.get(
            predict.remote(
                X_test_id,
                test_batch_size,
                self.model_cls,
                best_result.model_params,
                best_result.labels,
                best_checkpoint_id,
                best_result.checkpoint_name,
                master_ip,
                self.worker_gobbli_dir,
                self.worker_log_level,
                self.distributed,
            )
        )

        best_checkpoint = best_checkpoint_id
        if not is_ray_local_mode():
            best_checkpoint = ray.get(best_checkpoint_id)

        return ClassificationExperimentResults(
            X=X_test,
            labels=best_result.labels,  # type: ignore
            y_true=y_test,
            y_pred_proba=y_pred_proba,
            training_results=training_results,
            best_model_checkpoint=cast(Union[bytes, Path], best_checkpoint),
            best_model_checkpoint_name=best_result.checkpoint_name,
        )



# Copyright (c) Microsoft. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter

class LayerNorm(nn.Module):
    #ref: https://github.com/pytorch/pytorch/issues/1959
    #   :https://arxiv.org/pdf/1607.06450.pdf
    def __init__(self, hidden_size, eps=1e-4):
        super(LayerNorm, self).__init__()
        self.alpha = Parameter(torch.ones(1,1,hidden_size)) # gain g
        self.beta = Parameter(torch.zeros(1,1,hidden_size)) # bias b
        self.eps = eps

    def forward(self, x):
        """
        Args:
            :param x: batch * len * input_size

        Returns:
            normalized x
        """
        mu = torch.mean(x, 2, keepdim=True).expand_as(x)
        sigma = torch.std(x, 2, keepdim=True).expand_as(x)
        return (x - mu) / (sigma + self.eps) * self.alpha.expand_as(x) + self.beta.expand_as(x)



import pytest

from gobbli.model.use import USE


@pytest.mark.parametrize(
    "params,exception",
    [
        # Unknown param
        ({"unknown": None}, ValueError),
        # Bad value (use_model)
        ({"use_model": "bert"}, ValueError),
        # OK value (use_model)
        ({"use_model": "universal-sentence-encoder"}, None),
    ],
)
def test_init(params, exception):
    if exception is None:
        USE(**params)
    else:
        with pytest.raises(exception):
            USE(**params)



import logging
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from timeit import default_timer as timer
from typing import Any, Dict, Optional

import docker

from gobbli.model.context import ContainerTaskContext
from gobbli.util import (
    format_duration,
    generate_uuid,
    gobbli_version,
    is_dir_empty,
    model_dir,
    read_metadata,
    write_metadata,
)

LOGGER = logging.getLogger(__name__)


_WEIGHTS_DIR_NAME = "weights"


class BaseModel(ABC):
    """
    Abstract base class for all models.

    Derived classes should be careful to call super().__init__(...) with the appropriate
    arguments if they override __init__() to preserve all the functionality.

    Functionality to facilitate making GPU(s) available to derived classes is available.
    """

    # File containing information about the model, including type of model and gobbli version
    # the model was created under
    _INFO_FILENAME = "gobbli-model-info.json"

    # File containing model parameters (i.e. arguments to init())
    _METADATA_FILENAME = "gobbli-model-meta.json"

    _WEIGHTS_DIR_NAME = _WEIGHTS_DIR_NAME
    _CONTAINER_WEIGHTS_PATH = Path("/model") / _WEIGHTS_DIR_NAME

    def __init__(
        self,
        data_dir: Optional[Path] = None,
        load_existing: bool = False,
        use_gpu: bool = False,
        nvidia_visible_devices: str = "all",
        logger: Optional[logging.Logger] = None,
        **kwargs,
    ):
        """
        Create a model.

        Args:
          data_dir: Optional path to a directory used to store model data.  If not given,
            a unique directory under GOBBLI_DIR will be created and used.
          load_existing: If True, ``data_dir`` should be a directory that was previously used
            to create a model.  Parameters will be loaded to match the original model, and
            user-specified model parameters will be ignored.  If False, the data_dir must
            be empty if it already exists.
          use_gpu: If True, use the
            nvidia-docker runtime (https://github.com/NVIDIA/nvidia-docker) to expose
            NVIDIA GPU(s) to the container.  Will cause an error if the computer you're running
            on doesn't have an NVIDIA GPU and/or doesn't have the nvidia-docker runtime installed.
          nvidia_visible_devices: Which GPUs to make available to the container; ignored if
            ``use_gpu`` is False.  If not 'all', should be a comma-separated string: ex. ``1,2``.
          logger: If passed, use this logger for logging instead of the default module-level logger.
          **kwargs: Additional model-specific parameters to be passed to the model's :meth:`init` method.
        """
        self._logger = LOGGER
        if logger is not None:
            self._logger = logger

        if data_dir is None:
            self._data_dir = self.model_class_dir() / generate_uuid()
        else:
            self._data_dir = data_dir
        # Ensure we have an absolute data dir so any derived paths used in metadata files, etc
        # aren't ambiguous
        self._data_dir = self._data_dir.resolve()
        self._data_dir.mkdir(parents=True, exist_ok=True)

        class_name = self.__class__.__name__
        cur_gobbli_version = gobbli_version()

        if self.info_path.exists():
            info = read_metadata(self.info_path)
            if not info["class"] == class_name:
                raise ValueError(
                    f"Model class mismatch: the model stored in {data_dir} is of "
                    f"class '{info['class']}'.  Expected '{class_name}'."
                )
            if not info["gobbli_version"] == cur_gobbli_version:
                warnings.warn(
                    f"The model stored in {data_dir} was created with gobbli version "
                    f"{info['gobbli_version']}, but you're running version {cur_gobbli_version}. "
                    "You may encounter compatibility issues."
                )

        if load_existing and self.metadata_path.exists():
            params = read_metadata(self.metadata_path)
            if len(kwargs) > 0:
                warnings.warn(
                    "User-passed params ignored due to existing model being "
                    f"loaded: {kwargs}"
                )

        else:
            if not is_dir_empty(self._data_dir):
                raise ValueError(
                    f"data_dir '{self._data_dir}' is non-empty;"
                    " it must be empty to avoid overwriting data."
                )
            params = kwargs
            write_metadata(params, self.metadata_path)
            write_metadata(
                {"class": class_name, "gobbli_version": cur_gobbli_version},
                self.info_path,
            )

        self.use_gpu = use_gpu
        self.nvidia_visible_devices = nvidia_visible_devices

        self.docker_client = docker.from_env()

        self.init(params)

        self._logger.info(
            f"{class_name} initialized with data directory '{self._data_dir}'"
        )

    @property
    def logger(self) -> logging.Logger:
        """
        Returns:
          A logger for derived models to use.
        """
        return self._logger

    @property
    def info_path(self) -> Path:
        """
        Returns:
          The path to the model's info file, containing information about the model including
          the type of model, gobbli version it was trained using, etc.
        """
        return self.data_dir() / BaseModel._INFO_FILENAME

    @property
    def metadata_path(self) -> Path:
        """
        Returns:
         The path to the model's metadata file containing model-specific parameters.
        """
        return self.data_dir() / BaseModel._METADATA_FILENAME

    @abstractmethod
    def init(self, params: Dict[str, Any]):
        """
        Initialize a derived model using parameters specific to that model.

        Args:
          params: A dictionary where keys are parameter names and values are
            parameter values.
        """
        raise NotImplementedError

    def _base_docker_run_kwargs(self, context: ContainerTaskContext) -> Dict[str, Any]:
        """
        Establish a base set of docker run kwargs to handle GPU support, etc.
        Map directories as specified by the context.

        Returns:
            Base kwargs for any model that will be run using Docker.
        """
        kwargs = {
            "environment": {
                # Minimize the probability of containers exiting without dumping
                # buffered output
                "PYTHONUNBUFFERED": "1"
            },
            "detach": True,
            "volumes": {
                str(context.task_root_dir): {
                    "bind": str(context.container_root_dir),
                    "mode": "rw",
                },
                # Ideally we'd mount this as read-only, but some models (e.g. fastText)
                # need to write to their weights
                str(self.weights_dir): {
                    "bind": str(BaseModel._CONTAINER_WEIGHTS_PATH),
                    "mode": "rw",
                },
            },
        }  # type: Dict[str, Any]

        if self.use_gpu:
            kwargs["environment"][
                "NVIDIA_VISIBLE_DEVICES"
            ] = self.nvidia_visible_devices
            kwargs["runtime"] = "nvidia"

        return kwargs

    @property
    def _base_docker_build_kwargs(self) -> Dict[str, Any]:
        """
        Handle GPU support, etc via common args for any model Docker container.

        Returns:
            Base kwargs for any model that will be built using Docker.
        """
        kwargs = {"buildargs": {}}  # type: Dict[str, Any]

        if self.use_gpu:
            kwargs["buildargs"]["GPU"] = "1"

        return kwargs

    def data_dir(self) -> Path:
        """
        Returns:
            The main data directory unique to this instance of the model.
        """
        return self._data_dir

    @classmethod
    def model_class_dir(cls) -> Path:
        """
        Returns:
          A directory shared among all classes of the model.
        """
        return model_dir() / cls.__name__

    @property
    def class_weights_dir(self) -> Path:
        """
        The root directory used to store initial model weights (before fine-tuning).
        These should generally be some pretrained weights made available by model
        developers.  This directory will NOT be created by default; models should
        download their weights and remove the weights directory if the download doesn't
        finish properly.

        Most models making use of this directory will have multiple sets of weights and
        will need to store those in subdirectories under this directory.

        Returns:
            The path to the class-wide weights directory.
        """
        return self.model_class_dir() / BaseModel._WEIGHTS_DIR_NAME

    @property
    def weights_dir(self) -> Path:
        """
        The directory containing weights for a specific instance of the model.
        This is the class weights directory by default, but subclasses might
        define this property to return a subdirectory based on a set of pretrained
        model weights.

        Returns:
          The instance-specific weights directory.
        """
        return self.class_weights_dir

    def build(self):
        """
        Perform any pre-setup that needs to be done before running the model
        (building Docker images, etc).
        """
        self.logger.info("Starting build.")
        start = timer()
        self._build()
        end = timer()
        self.logger.info(f"Build finished in {format_duration(end - start)}.")

    @abstractmethod
    def _build(self):
        """
        Used for derived classes to define their implementation of the build method.
        """
        raise NotImplementedError



from pathlib import Path
from typing import Any, Callable, Dict, List

import click
import eli5
import numpy as np
import streamlit as st
from eli5.lime import TextExplainer

from gobbli.interactive.util import (
    DEFAULT_PREDICT_BATCH_SIZE,
    get_predictions,
    load_data,
    st_model_metadata,
    st_select_model_checkpoint,
)


def make_predict_func(
    model_cls: Any,
    model_kwargs: Dict[str, Any],
    unique_labels: List[str],
    checkpoint: str,
    batch_size: int,
) -> Callable[[List[str]], np.ndarray]:
    def predict(texts: List[str]) -> np.ndarray:
        preds = get_predictions(
            model_cls,
            model_kwargs,
            texts,
            unique_labels,
            checkpoint,
            batch_size=batch_size,
        ).values
        # Ensure rows sum to 1 -- eli5 raises an error if this isn't true
        preds = preds / preds.sum(axis=1, keepdims=1)
        return preds

    return predict


def st_lime_explanation(
    text: str,
    predict_func: Callable[[List[str]], np.ndarray],
    unique_labels: List[str],
    n_samples: int,
    position_dependent: bool = True,
):
    # TODO just use ELI5's built-in visualization when streamlit supports it:
    # https://github.com/streamlit/streamlit/issues/779
    with st.spinner("Generating LIME explanations..."):
        te = TextExplainer(
            random_state=1, n_samples=n_samples, position_dependent=position_dependent
        )
        te.fit(text, predict_func)
    st.json(te.metrics_)
    explanation = te.explain_prediction()
    explanation_df = eli5.format_as_dataframe(explanation)
    for target_ndx, target in enumerate(
        sorted(explanation.targets, key=lambda t: -t.proba)
    ):
        target_explanation_df = explanation_df[
            explanation_df["target"] == target_ndx
        ].copy()

        target_explanation_df["contribution"] = (
            target_explanation_df["weight"] * target_explanation_df["value"]
        )
        target_explanation_df["abs_contribution"] = abs(
            target_explanation_df["contribution"]
        )
        target_explanation_df = (
            target_explanation_df.drop("target", axis=1)
            .sort_values(by="abs_contribution", ascending=False)
            .reset_index(drop=True)
        )
        st.subheader(
            f"Target: {unique_labels[target_ndx]} (probability {target.proba:.4f}, score {target.score:.4f})"
        )
        st.dataframe(target_explanation_df)


@click.command()
@click.argument("model_data_dir", type=str)
@click.argument("data", type=str)
@click.option(
    "--n-rows",
    type=int,
    help="Number of rows to load from the data file.  If -1, load all rows.",
    default=-1,
    show_default=True,
)
@click.option(
    "--use-gpu/--use-cpu",
    default=False,
    help="Which device to run the model on. Defaults to CPU.",
)
@click.option(
    "--nvidia-visible-devices",
    default="all",
    help="Which GPUs to make available to the container; ignored if running on CPU. "
    "If not 'all', should be a comma-separated string: ex. ``1,2``.",
    show_default=True,
)
def run(
    model_data_dir: str,
    data: str,
    n_rows: int,
    use_gpu: bool,
    nvidia_visible_devices: str,
):
    st.sidebar.header("Model")
    model_data_path = Path(model_data_dir)
    model_cls, model_kwargs, checkpoint_meta = st_select_model_checkpoint(
        model_data_path, use_gpu, nvidia_visible_devices
    )
    st.title(f"Explaining: {model_cls.__name__} using {data}")

    model = model_cls(**model_kwargs)
    st_model_metadata(model)

    texts, labels = load_data(data, n_rows=None if n_rows == -1 else n_rows)

    st.sidebar.header("Example")
    example_ndx = st.sidebar.number_input(
        f"Index of Example (0 to {len(texts)-1})",
        min_value=0,
        max_value=len(texts) - 1,
        value=0,
    )

    if labels is None:
        header_text = "Example to Explain"
    else:
        header_text = f"Example to Explain (label: {labels[example_ndx]})"
    st.header(header_text)
    st.text(texts[example_ndx])

    st.sidebar.header("Predictions")
    predict_batch_size = st.sidebar.number_input(
        "Batch Size",
        min_value=1,
        max_value=len(texts),
        value=DEFAULT_PREDICT_BATCH_SIZE,
    )

    predict_func = make_predict_func(
        model_cls,
        model_kwargs,
        checkpoint_meta["labels"],
        checkpoint_meta["checkpoint"],
        predict_batch_size,
    )

    do_lime = st.sidebar.checkbox("Generate LIME explanation")

    do_run = st.sidebar.button("Run")

    if do_lime:
        st.sidebar.header("LIME Parameters")
        n_samples = st.sidebar.number_input(
            "Number of Samples", min_value=1, max_value=None, value=500
        )
        position_dependent = st.sidebar.checkbox(
            "Use position-dependent vectorizer", value=True
        )
        if do_run:
            st.header("LIME Explanation")
            st_lime_explanation(
                texts[example_ndx],
                predict_func,
                checkpoint_meta["labels"],
                n_samples,
                position_dependent=position_dependent,
            )

    if not do_lime:
        st.header("Explanation")
        st.markdown(
            "Select one or more explanation methods in the sidebar and click 'Run' to generate explanations."
        )


if __name__ == "__main__":
    # Streamlit doesn't like click exiting the script early after the function runs
    try:
        run()
    except SystemExit:
        pass



from typing import List, Optional

import numpy as np

import gobbli.io
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import EmbedMixin, TrainMixin


class RandomEmbedder(BaseModel, TrainMixin, EmbedMixin):
    """
    Dummy embeddings generator that returns random numbers as embeddings
    and has a stub training method to create a uniform API with other embedding
    models.

    Useful for ensuring user code works with the gobbli input/output format
    without having to build a time-consuming model.
    """

    SEED = 1234
    DIMENSIONALITY = 32

    def init(self, params):
        pass

    def _build(self):
        """
        No build step required for this model.
        """

    def tokenize(self, X: List[str]) -> List[List[str]]:
        """
        Return a tokenized list of documents.
        """
        return [x.split() for x in X]

    def _train(
        self, train_input: gobbli.io.TrainInput, context: ContainerTaskContext
    ) -> gobbli.io.TrainOutput:
        """
        No training needed -- stubbed for API uniformity.
        """
        return gobbli.io.TrainOutput(
            valid_loss=0, valid_accuracy=0, train_loss=0, labels=[]
        )

    def _embed(
        self, embed_input: gobbli.io.EmbedInput, context: ContainerTaskContext
    ) -> gobbli.io.EmbedOutput:
        """
        Generate random embeddings.
        """
        np.random.seed(RandomEmbedder.SEED)
        X_tokenized = self.tokenize(embed_input.X)

        embeddings = []
        for tokens in X_tokenized:
            token_embeddings = np.random.rand(
                # sequence length
                len(tokens),
                # dimensionality
                RandomEmbedder.DIMENSIONALITY,
            )

            if embed_input.pooling == gobbli.io.EmbedPooling.MEAN:
                token_embeddings = np.mean(token_embeddings, axis=0)

            embeddings.append(token_embeddings)

        embed_tokens = None  # type: Optional[List[List[str]]]
        if embed_input.pooling == gobbli.io.EmbedPooling.NONE:
            embed_tokens = X_tokenized[:]

        return gobbli.io.EmbedOutput(X_embedded=embeddings, embed_tokens=embed_tokens)



import csv
import datetime as dt
import inspect
import itertools
import json
import os
import random
from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, TypeVar

import pandas as pd
import streamlit as st

import gobbli
import gobbli.model
from gobbli.dataset.base import BaseDataset
from gobbli.io import PredictInput, TaskIO
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import TrainMixin
from gobbli.util import read_metadata, truncate_text

DEFAULT_PREDICT_BATCH_SIZE = PredictInput.predict_batch_size


@st.cache
def get_label_indices(labels: List[str]) -> Dict[str, List[int]]:
    label_indices = defaultdict(list)
    for i, label in enumerate(labels):
        label_indices[label].append(i)
    return label_indices


def _read_delimited(
    data_file: Path, delimiter: str, n_rows: Optional[int] = None
) -> Tuple[List[str], Optional[List[str]]]:
    """
    Read up to n_rows lines from the given delimited text file and return lists
    of the texts and labels.  Texts must be stored in a column named "text", and
    labels (if any) must be stored in a column named "label".

    Args:
      data_file: Data file containing one text per line.
      delimiter: Field delimiter for the data file.
      n_rows: The maximum number of rows to read.

    Returns:
      2-tuple: list of read texts and corresponding list of read labels.
    """
    with open(data_file, "r") as f:
        reader = csv.DictReader(f, delimiter=delimiter)
        fieldnames = set(reader.fieldnames)

        if "text" not in fieldnames:
            raise ValueError("Delimited text file doesn't contain a 'text' column.")
        has_labels = "label" in fieldnames

        rows = list(itertools.islice(reader, n_rows))

    texts: List[str] = []
    labels: List[str] = []

    for row in rows:
        texts.append(row["text"])
        if has_labels:
            labels.append(row["label"])

    return texts, labels if has_labels else None


def _read_lines(data_file: Path, n_rows: Optional[int] = None) -> List[str]:
    """
    Read up to n_rows lines from the given text file and return them in a list.

    Args:
      data_file: Data file containing one text per line.
      n_rows: The maximum number of rows to read.

    Returns:
      List of read lines.
    """
    with open(data_file, "r") as f:
        return list(itertools.islice((l.strip() for l in f), n_rows))


def read_data_file(
    data_file: Path, n_rows: Optional[int] = None
) -> Tuple[List[str], Optional[List[str]]]:
    """
    Read data to explore from a file.  Rows may be sampled using the n_rows argument.

    Args:
      data_file: Path to a data file to read.
      n_rows: The maximum number of rows to read.

    Returns:
      2-tuple: list of read texts and a list of read labels (if any)
    """
    extension = data_file.suffix
    if extension == ".tsv":
        texts, labels = _read_delimited(data_file, "\t", n_rows=n_rows)
    elif extension == ".csv":
        texts, labels = _read_delimited(data_file, ",", n_rows=n_rows)
    elif extension == ".txt":
        labels = None
        texts = _read_lines(data_file, n_rows=n_rows)
    else:
        raise ValueError(f"Data file extension '{extension}' is unsupported.")

    return texts, labels


def sample_dataset(
    dataset: BaseDataset, n_rows: Optional[int] = None
) -> Tuple[List[str], List[str]]:
    """
    Sample the given number of rows from the given dataset.

    Args:
      dataset: Loaded dataset to sample from.
      n_rows: Optional number of rows to sample.  If None, return all rows.

    Returns:
      2-tuple: a list of texts and a list of labels.  If n_rows was given, these will
      be no longer than n_rows.
    """
    # Apply limit to the dataset, if any
    if n_rows is None:
        texts = dataset.X_train() + dataset.X_test()
        labels = dataset.y_train() + dataset.y_test()
    else:
        # Try to reach the limit from the train split only first
        train_texts = dataset.X_train()[:n_rows]
        train_labels = dataset.y_train()[:n_rows]

        if len(train_texts) < n_rows:
            # If we need more rows to reach the limit, get them
            # from the test set
            test_limit = n_rows - len(train_texts)
            test_texts = dataset.X_test()[:test_limit]
            test_labels = dataset.y_test()[:test_limit]

            texts = train_texts + test_texts
            labels = train_labels + test_labels
        else:
            # Otherwise, just use the limited train data
            texts = train_texts
            labels = train_labels

    return texts, labels


@st.cache(show_spinner=True)
def read_data_file_cached(
    # Streamlit errors sometimes when hashing Path objects, so use a string.
    # https://github.com/streamlit/streamlit/issues/857
    data_file: str,
    n_rows: Optional[int] = None,
) -> Tuple[List[str], Optional[List[str]]]:
    """
    Streamlit-cached wrapper around :func:`read_data_file` for performance.
    """
    return read_data_file(Path(data_file), n_rows=n_rows)


def load_data(
    data: str, n_rows: Optional[int]
) -> Tuple[List[str], Optional[List[str]]]:
    """
    Load data according to the given 'data' string and row limit.

    Args:
      data: Could be either the name of a gobbli dataset class or a path
      to a data file in a supported format.
      n_rows: Optional limit on number of rows read from the data.

    Returns:
      2-tuple: List of texts and list of labels.
    """
    if os.path.exists(data):
        data_path = Path(data)
        texts, labels = read_data_file_cached(
            str(data_path), n_rows=None if n_rows == -1 else n_rows
        )
    elif data in gobbli.dataset.__all__:
        dataset = getattr(gobbli.dataset, data).load()
        texts, labels = sample_dataset(dataset, None if n_rows == -1 else n_rows)
    else:
        raise ValueError(
            "data argument did not correspond to an existing data file in a "
            "supported format or a built-in gobbli dataset.  Available datasets: "
            f"{gobbli.dataset.__all__}"
        )

    return texts, labels


T = TypeVar("T")


@st.cache
def safe_sample(l: Sequence[T], n: int, seed: Optional[int] = None) -> List[T]:
    if seed is not None:
        random.seed(seed)

    # Prevent an error from trying to sample more than the population
    return list(random.sample(l, min(n, len(l))))


def st_sample_data(
    texts: List[str], labels: Optional[List[str]]
) -> Tuple[List[str], Optional[List[str]]]:
    """
    Generate streamlit sidebar widgets to facilitate sampling a dataset at runtime.

    Args:
      texts: Full list of texts to sample from.
      labels: Full list of labels to sample from.

    Returns:
      2-tuple: the list of sampled texts and list of sampled labels.
    """
    st.sidebar.header("Sample")

    if st.sidebar.button("Randomize Seed"):
        default_seed = random.randint(0, 1000000)
    else:
        default_seed = 1
    sample_seed = st.sidebar.number_input("Sample Seed", value=default_seed)

    sample_size = st.sidebar.slider(
        "Sample Size", min_value=1, max_value=len(texts), value=min(100, len(texts))
    )

    sample_indices = safe_sample(range(len(texts)), sample_size, seed=sample_seed)

    sampled_texts = [texts[i] for i in sample_indices]

    if labels is None:
        sampled_labels = None
    else:
        sampled_labels = [labels[i] for i in sample_indices]

    return sampled_texts, sampled_labels


def st_example_documents(
    texts: List[str], labels: Optional[List[str]], truncate_len: int
):
    """
    Generate streamlit elements showing example documents (and optionally labe
    """
    df = pd.DataFrame({"Document": [truncate_text(t, truncate_len) for t in texts]})
    if labels is not None:
        df["Label"] = labels
    st.table(df)


def format_task(task_dir: Path) -> str:
    """
    Format the given task for a human-readable dropdown.

    Args:
      task_dir: Directory where the task's data is stored.

    Returns:
      String-formatted, human-readable task metadata.
    """
    task_id = task_dir.name
    task_creation_time = dt.datetime.fromtimestamp(task_dir.stat().st_birthtime)
    return f"{task_id[:5]} - Created {task_creation_time.strftime('%Y-%m-%d %H:%M:%S')}"


def st_select_model_checkpoint(
    model_data_path: Path, use_gpu: bool, nvidia_visible_devices: str
) -> Tuple[Any, Dict[str, Any], Dict[str, Any]]:
    """
    Generate widgets allowing for users to select a checkpoint from a given model directory.

    Args:
      model_data_path: Path to the model data directory to search for checkpoints.
      use_gpu: If True, initialize the model using a GPU.
      nvidia_visible_devices: The list of devices to make available to the model container.
       Should be either "all" or a comma-separated list of device IDs (ex "1,2").

    Returns:
      A 3-tuple: the class of model corresponding to the checkpoint, the kwargs to initialize
      the model with, and the metadata for the checkpoint.
    """
    try:
        model_info = read_metadata(model_data_path / BaseModel._INFO_FILENAME)
    except FileNotFoundError:
        raise ValueError(
            "The passed model data directory does not appear to contain a saved gobbli model. "
            "Did you pass the right directory?"
        )

    model_cls_name = model_info["class"]
    if not hasattr(gobbli.model, model_cls_name):
        raise ValueError(f"Unknown model type: {model_cls_name}")
    model_cls = getattr(gobbli.model, model_cls_name)

    model_kwargs = {
        "data_dir": model_data_path,
        "load_existing": True,
        "use_gpu": use_gpu,
        "nvidia_visible_devices": nvidia_visible_devices,
    }
    model = model_cls(**model_kwargs)

    task_metadata = {}
    if isinstance(model, TrainMixin):
        # The model can be trained, so it may have some trained weights
        model_train_dir = model.train_dir()

        # See if any checkpoints are available for the given model
        for task_dir in model_train_dir.iterdir():
            task_context = ContainerTaskContext(task_dir)
            output_dir = task_context.host_output_dir

            if output_dir.exists():
                metadata_path = output_dir / TaskIO._METADATA_FILENAME
                if metadata_path.exists():
                    with open(metadata_path, "r") as f:
                        metadata = json.load(f)

                        if "checkpoint" in metadata:
                            task_formatted = format_task(task_dir)
                            task_metadata[task_formatted] = metadata

    if len(task_metadata) == 0:
        st.error("No trained checkpoints found for the given model.")
        return

    model_checkpoint = st.sidebar.selectbox(
        "Model Checkpoint", list(task_metadata.keys())
    )
    return model_cls, model_kwargs, task_metadata[model_checkpoint]


def st_select_untrained_model(
    use_gpu: bool,
    nvidia_visible_devices: str,
    predicate: Callable[[Any], bool] = lambda _: True,
) -> Optional[Tuple[Any, Dict[str, Any]]]:
    """
    Generate widgets allowing users to select an untrained model and apply arbitrary
    model parameters.

    Args:
      use_gpu: If True, initialize the model using a GPU.
      nvidia_visible_devices: The list of devices to make available to the model container.
       Should be either "all" or a comma-separated list of device IDs (ex "1,2").
      predicate: A predicate used to filter the avaliable model classes.

    Returns:
      A 2-tuple: the class of model and the kwargs to initialized the model with.
    """
    model_choices = [
        cls.__name__
        for name, cls in inspect.getmembers(gobbli.model)
        if inspect.isclass(cls) and issubclass(cls, BaseModel) and predicate(cls)
    ]

    model_cls_name = st.sidebar.selectbox("Model Class", model_choices)
    model_params_str = st.sidebar.text_area("Model Parameters (JSON)", value="{}")

    # Slight convenience if the user deletes the text area contents
    if model_params_str == "":
        model_params_str = "{}"

    model_cls = getattr(gobbli.model, model_cls_name)

    # Validate the model parameter JSON
    try:
        model_params = json.loads(model_params_str)
    except Exception:
        st.sidebar.error("Model parameters must be valid JSON.")
        return None

    model_kwargs = {
        "use_gpu": use_gpu,
        "nvidia_visible_devices": nvidia_visible_devices,
        **model_params,
    }

    # Validate the parameters using the model initialization function
    try:
        model_cls(**model_kwargs)
    except (TypeError, ValueError) as e:
        st.sidebar.error(f"Error validating model parameters: {e}")
        return None

    return model_cls, model_kwargs


def st_model_metadata(model: BaseModel):
    with open(model.metadata_path, "r") as f:
        model_metadata = json.load(f)

    st.header("Model Metadata")
    st.json(model_metadata)


@st.cache(show_spinner=True)
def get_predictions(
    model_cls: BaseModel,
    model_kwargs: Dict[str, Any],
    texts: List[str],
    unique_labels: List[str],
    checkpoint: str,
    batch_size: int = DEFAULT_PREDICT_BATCH_SIZE,
) -> pd.DataFrame:
    """
    Run the given model on the given texts and return the probabilities.

    Args:
      model: Model to use for prediction.
      texts: List of texts to generate predictions for.
      unique_labels: Ordered list of unique labels.
      checkpoint: Model checkpoint to use for prediction.
      batch_size: Batch size for prediction.

    Returns:
      A dataframe containing the predicted probability for each text and label.
    """
    predict_input = PredictInput(
        X=texts,
        labels=unique_labels,
        checkpoint=checkpoint,
        predict_batch_size=batch_size,
    )
    model = model_cls(**model_kwargs)
    model.build()
    predict_output = model.predict(predict_input)
    return predict_output.y_pred_proba



import argparse
import itertools
import json
from pathlib import Path

import numpy as np
import pandas as pd

import torch
import transformers
from transformers import AdamW, get_linear_schedule_with_warmup


def read_unique_labels(labels_path):
    """
    Read the list of unique labels from the given path.
    """
    return labels_path.read_text().split("\n")


def read_data(file_path, has_labels):
    """
    Read some text data with optional labels from the given path.
    Return a 2-tuple of texts and labels (if any).
    """
    df = pd.read_csv(file_path, sep="\t", dtype="str", keep_default_na=False)
    X = df["Text"].tolist()
    y = None
    if has_labels:
        y = df["Label"].tolist()
    return X, y


def batch_gen(l, batch_size):
    for i in range(0, len(l), batch_size):
        yield l[i : i + batch_size]


def encode_batch(batch, tokenizer, max_seq_length, return_tokens=False):
    """
    Tokenize and encode the given input; stack it in a batch with fixed length
    according to the longest sequence in the batch.

    Return a tuple.  If asked to return tokens, return the encoded batch and a nested
    list of the tokenized texts.  Otherwise, just return the batch.
    """
    # Truncate input based on the sequence length
    # before stacking as a batch
    # Also return a boolean array indicating which tokens are masked
    encoded_texts = []

    for text in batch:
        encoded_text = torch.tensor(tokenizer.encode(text))[:max_seq_length]
        encoded_texts.append(encoded_text)

    encoded_batch = torch.nn.utils.rnn.pad_sequence(encoded_texts, batch_first=True)

    if return_tokens:
        tokens = [tokenizer.tokenize(text) for text in batch]
        return encoded_batch, tokens
    else:
        return (encoded_batch,)


def encode_labels(y, labels):
    """
    Convert string labels to autoincrementing integers.
    """
    mapping = {l: i for i, l in enumerate(labels)}
    return torch.as_tensor([mapping[l] for l in y])


def decode_labels(y_encoded, labels):
    """
    Convert autoincrementing integers back to strings.
    """
    return [labels[i] for i in y_encoded]


def tsv_to_encoded_batches(
    input_path, tokenizer, labels, batch_size, max_seq_length, return_tokens=False
):
    """
    Convert data in a csv file with optional labels to a list
    of encoded batches.  Use a generator to guard against memory errors.

    If `return_tokens` is True, add a nested list containing the tokenized
    text for each row.

    Just return encoded texts and optional tokens if no labels,
    else also return encoded labels.
    """
    has_labels = labels is not None
    X, y = read_data(input_path, has_labels)

    if y is None:
        y_batch_gen = itertools.repeat(None)
    else:
        y_batch_gen = batch_gen(y, batch_size)

    for X_batch, y_batch in zip(batch_gen(X, batch_size), y_batch_gen):
        X_encoded_batch = encode_batch(
            X_batch, tokenizer, max_seq_length, return_tokens=return_tokens
        )

        if has_labels:
            y_encoded_batch = encode_labels(y_batch, labels)
            yield (*X_encoded_batch, y_encoded_batch)
        else:
            yield X_encoded_batch


def num_batches(data_path, batch_size):
    """
    Count the number of batches contained in the given TSV data file assuming
    the given batch size.
    """
    X, _ = read_data(data_path, False)
    return (len(X) - 1) // batch_size + 1


def train(
    *,
    input_dir,
    output_dir,
    model,
    config,
    tokenizer,
    train_batch_size,
    valid_batch_size,
    num_train_epochs,
    device,
    n_gpu,
    max_seq_length,
    labels,
    lr,
    adam_eps,
    gradient_accumulation_steps,
):
    """
    Train the passed model on the given data.  Return training/validation metrics
    and save the model to the specified output directory.
    """
    # Prepare optimizer and scheduler
    # Adapted from https://github.com/huggingface/transformers/blob/master/examples/run_squad.py
    optimizer = AdamW(model.parameters(), lr=lr, eps=adam_eps)
    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=0,
        num_training_steps=num_train_epochs
        * num_batches(input_dir / "train.tsv", train_batch_size)
        // gradient_accumulation_steps,
    )

    model.zero_grad()

    for epoch in range(num_train_epochs):
        print(f"Start epoch {epoch}")
        model.train()
        train_loss = 0.0
        train_count = 0

        for i, (X, y) in enumerate(
            tsv_to_encoded_batches(
                input_dir / "train.tsv",
                tokenizer,
                labels,
                train_batch_size,
                max_seq_length,
            )
        ):
            X = X.to(device)
            y = y.to(device)
            outputs = model(input_ids=X, labels=y)
            loss = outputs[0]
            if gradient_accumulation_steps > 1:
                loss /= gradient_accumulation_steps

            if n_gpu > 1:
                loss = loss.mean()  # average loss on parallel training

            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)

            train_loss += loss.detach().item()
            train_count += X.size(0)

            if (i + 1) % gradient_accumulation_steps == 0:
                optimizer.step()
                scheduler.step()
                model.zero_grad()

        print(f"Epoch {epoch} train loss: {train_loss / train_count}")

        model.eval()
        valid_loss = 0.0
        valid_count = 0
        valid_correct = 0

        for X, y in tsv_to_encoded_batches(
            input_dir / "dev.tsv", tokenizer, labels, valid_batch_size, max_seq_length
        ):
            X = X.to(device)
            y = y.to(device)

            with torch.no_grad():
                outputs = model(input_ids=X, labels=y)
                loss = outputs[0]
                logits = outputs[1]
                # Keep preds on the device rather than moving to CPU, since we'll compare to y,
                # which is on the device
                preds = torch.argmax(logits.detach(), 1)

            if n_gpu > 1:
                loss = loss.mean()  # average loss on parallel training

            valid_loss += loss.item()
            valid_count += X.size(0)
            valid_correct += (y == preds).sum().item()

        print(f"Epoch {epoch} validation loss: {valid_loss / valid_count}")

    checkpoint_dir = output_dir / "checkpoint"
    checkpoint_dir.mkdir(exist_ok=True, parents=True)

    if hasattr(model, "module"):
        # DataParallel object -- unpack the module before saving
        model.module.save_pretrained(checkpoint_dir)
    else:
        # Plain transformers model
        model.save_pretrained(checkpoint_dir)

    tokenizer.save_pretrained(checkpoint_dir)
    config.save_pretrained(checkpoint_dir)

    return {
        "mean_train_loss": train_loss / train_count,
        "mean_valid_loss": valid_loss / valid_count,
        "valid_accuracy": valid_correct / valid_count,
    }


def predict(
    *,
    input_dir,
    output_dir,
    model,
    tokenizer,
    predict_batch_size,
    device,
    max_seq_length,
    labels,
):

    pred_probas = []

    for (X,) in tsv_to_encoded_batches(
        input_dir / "test.tsv", tokenizer, None, predict_batch_size, max_seq_length
    ):
        X = X.to(device)

        with torch.no_grad():
            outputs = model(input_ids=X)
            logits = outputs[0]
            pred_probas.append(torch.softmax(logits.detach().cpu(), 1).numpy())

    df = pd.DataFrame(np.vstack(pred_probas))
    df.columns = labels
    df.to_csv(output_dir / "test_results.tsv", index=False, sep="\t")


def embed(
    *,
    input_dir,
    output_dir,
    model,
    tokenizer,
    embed_batch_size,
    device,
    max_seq_length,
    embed_pooling,
    embed_layer,
):
    embeddings = []

    with open(output_dir / "embeddings.jsonl", "w") as f:
        for X, tokens in tsv_to_encoded_batches(
            input_dir / "input.tsv",
            tokenizer,
            None,
            embed_batch_size,
            max_seq_length,
            return_tokens=True,
        ):
            X = X.to(device)

            with torch.no_grad():
                outputs = model(input_ids=X)
                hidden_states = outputs[-1]
                hidden_state = hidden_states[embed_layer]

                if embed_pooling == "mean":
                    embeddings = hidden_state.mean(dim=1).cpu().numpy().tolist()
                    for row in embeddings:
                        row_json = {"embedding": row}
                        f.write(f"{json.dumps(row_json)}\n")
                elif embed_pooling == "none":
                    embeddings = hidden_state.cpu().numpy().tolist()
                    for row, tokens in zip(embeddings, tokens):
                        # Truncate padding embeddings if needed
                        row_json = {"embedding": row[: len(tokens)], "tokens": tokens}
                        f.write(f"{json.dumps(row_json)}\n")
                else:
                    raise ValueError(f"Unsupported pooling type: {embed_pooling}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "mode",
        choices=["train", "predict", "embed"],
        help="Action to perform with the model. Train: read in a training "
        "and validation dataset, train on training and evaluate the model "
        "on validation dataset. Predict: output predictions on a test dataset. "
        "Embed: return embeddings for a test dataset.",
    )
    parser.add_argument(
        "--input-dir",
        required=True,
        help="Directory containing input files.  The exact files needed vary "
        "depending on the mode.",
    )
    parser.add_argument(
        "--output-dir",
        required=True,
        help="Directory containing output files.  The exact files created vary "
        "depending on the mode.",
    )
    parser.add_argument(
        "--cache-dir",
        required=True,
        help="Directory to use for caching transformers downloads.",
    )
    parser.add_argument(
        "--config-overrides",
        help="Path to a JSON file containing config overrides for the given model.",
    )
    parser.add_argument(
        "--model",
        required=True,
        help="Name of the transformers model to use.  If `mode` is 'train' or 'predict', "
        "this value should ensure `from transformers import <model>ModelForSequenceClassification` "
        "is a valid import. For example, model = 'Bert' -> from transformers import "
        "BertForSequenceClassification.  If `mode` is 'embed', the import will be `<model>Model`.",
    )
    parser.add_argument(
        "--weights",
        required=True,
        help="Name of the pretrained weights to use or path to a file containing "
        "pretrained weights.  If a name, must correspond to a valid weights name "
        "for the chosen model.",
    )
    parser.add_argument(
        "--train-batch-size",
        type=int,
        default=32,
        help="Per-GPU batch size to use for training.",
    )
    parser.add_argument(
        "--valid-batch-size",
        type=int,
        default=32,
        help="Per-GPU batch size to use for validation.",
    )
    parser.add_argument(
        "--predict-batch-size",
        type=int,
        default=32,
        help="Per-GPU batch size to use for prediction.",
    )
    parser.add_argument(
        "--embed-batch-size",
        type=int,
        default=32,
        help="Per-GPU batch size to use for embedding.",
    )
    parser.add_argument(
        "--num-train-epochs",
        type=int,
        default=3,
        help="Number of epochs to run training on.",
    )
    parser.add_argument(
        "--max-seq-length",
        type=int,
        default=128,
        help="Truncate sequences to this length after tokenization.  If set too large for "
        "the model used, this may cause errors.",
    )
    parser.add_argument(
        "--lr", type=float, default=5e-5, help="Learning rate for the AdamW optimizer."
    )
    parser.add_argument(
        "--adam-eps",
        type=float,
        default=1e-8,
        help="Epsilon value for the AdamW optimizer.",
    )
    parser.add_argument(
        "--gradient-accumulation-steps",
        type=int,
        default=1,
        help="Number of training steps to accumulate gradients before updating model weights.",
    )
    parser.add_argument(
        "--embed-pooling",
        choices=["mean", "none"],
        default="mean",
        help="Pooling strategy to use for combining embeddings.  If None, embeddings for "
        "each token will be returned along with the token.  Ignored if method != 'embed'.",
    )
    parser.add_argument(
        "--embed-layer",
        type=int,
        default=-2,
        help="Index of the model layer to use for generating embeddings.",
    )

    args = parser.parse_args()

    if args.gradient_accumulation_steps <= 0:
        raise ValueError(
            f"Gradient accumulation steps must be >0, got '{args.gradient_accumulation_steps}'"
        )

    input_dir = Path(args.input_dir)
    output_dir = Path(args.output_dir)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    n_gpu = torch.cuda.device_count()

    print(f"Using device: {device}")
    print(f"Number of GPUs: {n_gpu}")

    # Which model class we import depends on the task
    if args.mode in ("train", "predict"):
        model_name = f"{args.model}ForSequenceClassification"
    elif args.mode == "embed":
        model_name = f"{args.model}Model"

    tokenizer_name = f"{args.model}Tokenizer"
    config_name = f"{args.model}Config"

    print("Initializing transformer...")
    print(f"  Model: {model_name}")
    print(f"    Weights: {args.weights}")
    print(f"  Tokenizer: {tokenizer_name}")
    print(f"  Config: {config_name}")

    model_cls = getattr(transformers, model_name)
    tokenizer_cls = getattr(transformers, tokenizer_name)
    config_cls = getattr(transformers, config_name)

    tokenizer = tokenizer_cls.from_pretrained(args.weights, cache_dir=args.cache_dir)
    if tokenizer is None:
        raise ValueError("Failed to acquire tokenizer")
    config = config_cls.from_pretrained(args.weights, cache_dir=args.cache_dir)
    if config is None:
        raise ValueError("Failed to acquire config")

    # Apply config overrides specified by the user
    if args.config_overrides is not None:
        with open(args.config_overrides, "r") as f:
            config_overrides = json.load(f)
            for key, val in config_overrides.items():
                print(
                    f"Overriding config key '{key}' with value '{val}' "
                    f"(old value: '{getattr(config, key)}"
                )
                setattr(config, key, val)

    # Set the number of labels on the config object appropriately before we
    # initialize the model using the config
    # We need a list of labels for training or prediction so we know what the
    # output shape of the model should be
    labels = None
    if args.mode in ("train", "predict"):
        labels_file = input_dir / "labels.tsv"
        labels = read_unique_labels(input_dir / "labels.tsv")
        config.num_labels = len(labels)
        if config.num_labels == 1:
            raise ValueError(
                "transformers calculates regression loss when only one "
                "label is given, so it doesn't support classification when only "
                "a single label is available."
            )

        print(f"Inferred number of labels: {config.num_labels}")

    # We need to set the config to output hidden states if we're generating embeddings
    # DON'T output attentions so we can rely on hidden states being the last value in
    # the model output
    if args.mode == "embed":
        config.output_hidden_states = True
        config.output_attentions = False
        print("Outputting hidden states for embeddings")

    model = model_cls.from_pretrained(
        args.weights, cache_dir=args.cache_dir, config=config
    )
    if model is None:
        raise ValueError("Failed to acquire model")

    # If we have multiple GPUs, size the batches appropriately so DataParallel
    # distributes them correctly
    batch_multiplier = max(1, n_gpu)
    train_batch_size = args.train_batch_size * batch_multiplier
    valid_batch_size = args.valid_batch_size * batch_multiplier
    predict_batch_size = args.predict_batch_size * batch_multiplier

    model.to(device)
    if n_gpu > 1:
        model = torch.nn.DataParallel(model)

    if args.mode == "train":
        metrics = train(
            input_dir=input_dir,
            output_dir=output_dir,
            model=model,
            config=config,
            tokenizer=tokenizer,
            train_batch_size=train_batch_size,
            valid_batch_size=valid_batch_size,
            num_train_epochs=args.num_train_epochs,
            device=device,
            n_gpu=n_gpu,
            max_seq_length=args.max_seq_length,
            labels=labels,
            lr=args.lr,
            adam_eps=args.adam_eps,
            gradient_accumulation_steps=args.gradient_accumulation_steps,
        )
        with open(output_dir / "valid_results.json", "w") as f:
            json.dump(metrics, f)

    elif args.mode == "predict":
        predict(
            input_dir=input_dir,
            output_dir=output_dir,
            model=model,
            tokenizer=tokenizer,
            predict_batch_size=args.predict_batch_size,
            device=device,
            max_seq_length=args.max_seq_length,
            labels=labels,
        )
    elif args.mode == "embed":
        embed(
            input_dir=input_dir,
            output_dir=output_dir,
            model=model,
            tokenizer=tokenizer,
            embed_batch_size=args.embed_batch_size,
            device=device,
            max_seq_length=args.max_seq_length,
            embed_pooling=args.embed_pooling,
            embed_layer=args.embed_layer,
        )
    else:
        raise ValueError(f"invalid mode: {args.mode}")



import warnings

import pkg_resources

import gobbli.augment as augment
import gobbli.dataset as dataset
import gobbli.experiment as experiment
import gobbli.io as io
import gobbli.model as model
from gobbli.util import TokenizeMethod

# Warn the user of potential conflicts using the old third-party typing/dataclasses
# modules
for conflicting_pkg in ("typing", "dataclasses"):
    req = pkg_resources.Requirement.parse(conflicting_pkg)
    if pkg_resources.working_set.find(req) is not None:
        warnings.warn(
            f"You've installed a third-party module named '{conflicting_pkg}' which "
            "conflicts with a standard library module of the same name.  This can cause "
            "errors when unpickling code, e.g. when running experiments using Ray. Consider "
            f"uninstalling the module:\n\npip uninstall {conflicting_pkg}"
        )

__all__ = [
    # Modules
    "augment",
    "dataset",
    "experiment",
    "model",
    "io",
    # Misc top level imports
    "TokenizeMethod",
]



from pathlib import Path

import pytest
import ray

from gobbli.test.util import MockDataset, MockExperiment, MockModel, skip_if_no_gpu


def test_base_experiment_init(tmpdir):
    tmpdir_path = Path(tmpdir)
    ds = MockDataset.load()

    # Create experiment
    exp = MockExperiment(MockModel, ds, data_dir=tmpdir_path / "test")
    assert exp.metadata_path.exists()

    # Shouldn't be able to create another experiment without ignoring the ray initialization error
    with pytest.raises(RuntimeError):
        MockExperiment(MockModel, ds, data_dir=tmpdir_path / "test2")
    MockExperiment(
        MockModel, ds, data_dir=tmpdir_path / "test3", ignore_ray_initialized_error=True
    )

    # Shouldn't be able to create another experiment in the same path
    with pytest.raises(ValueError):
        MockExperiment(
            MockModel,
            ds,
            data_dir=tmpdir_path / "test",
            ignore_ray_initialized_error=True,
        )

    # ...unless we pass overwrite_existing = True
    MockExperiment(
        MockModel,
        ds,
        data_dir=tmpdir_path / "test",
        ignore_ray_initialized_error=True,
        overwrite_existing=True,
    )

    # Limit should be obeyed
    assert len(exp.X) == len(MockDataset.X_TRAIN_VALID) + len(MockDataset.X_TEST)
    assert len(exp.y) == len(MockDataset.Y_TRAIN_VALID) + len(MockDataset.Y_TEST)

    exp_limit = MockExperiment(
        MockModel,
        ds,
        limit=1,
        data_dir=tmpdir_path / "test_limit",
        ignore_ray_initialized_error=True,
    )
    assert len(exp_limit.X) == 1
    assert len(exp_limit.y) == 1


def test_base_experiment_gpu(tmpdir, request):
    skip_if_no_gpu(request.config)

    tmpdir_path = Path(tmpdir)
    ds = MockDataset.load()

    MockExperiment(
        MockModel,
        ds,
        data_dir=tmpdir_path / "test",
        ray_kwargs={"num_gpus": 1},
        ignore_ray_initialized_error=True,
    )

    # Make sure GPUs are available
    # in a mock remote function
    # They won't necessarily be available on the master process
    @ray.remote(num_gpus=1)
    def find_gpus():
        return ray.get_gpu_ids()

    assert len(ray.get(find_gpus.remote())) > 0



import enum
import tempfile
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from gobbli.util import TokenizeMethod, detokenize, pred_prob_to_pred_label, tokenize


def _check_string_list(obj: Any):
    """
    Verify a given object is a list containing strings.
    """
    if not isinstance(obj, list):
        raise TypeError(f"obj must be a list, got '{type(obj)}'")
    if len(obj) > 0 and not isinstance(obj[0], str):
        raise TypeError(f"obj must contain strings, got '{type(obj[0])}'")


def validate_X(X: List[str]):
    """
    Confirm a given array matches the expected type for input.

    Args:
      X: Something that should be valid model input.
    """
    _check_string_list(X)


def validate_y(y: List[str]):
    """
    Confirm a given array matches the expected type and length for model output
    (expected or predicted).

    Args:
      y: Something that should be valid model output.
    """
    _check_string_list(y)


def validate_X_y(X: List[str], y: List[Any]):
    """
    Assuming X is valid input and y is valid output, ensure they match sizes.

    Args:
      X: Something that should be valid model input.
      y: Something that should be valid model output.
    """
    if len(X) != len(y):
        raise ValueError(
            f"X and y must have the same length; X has length {len(X)}, and y has length {len(y)}"
        )


class TaskIO(ABC):
    """
    Base class for classes used for task input/output.
    """

    _METADATA_FILENAME = "gobbli-task-meta.json"

    @abstractmethod
    def metadata(self) -> Dict[str, Any]:
        """
        Returns:
          The task information that constitutes its metadata --
          generally parameters of an input task and/or summarized results
          of an output task.
        """
        raise NotImplementedError


@dataclass
class TrainInput(TaskIO):
    """
    Input for training a model.  See :meth:`gobbli.model.mixin.TrainMixin.train`.

    Args:
      X_train: Documents used for training.
      y_train: Labels for training documents.
      X_valid: Documents used for validation.
      y_valid: Labels for validation documents.
      train_batch_size: Number of observations per batch on the training dataset.
      valid_batch_size: Number of observations per batch on the validation dataset.
      num_train_epochs: Number of epochs to use for training.
      checkpoint: Checkpoint containing trained weights for the model.  If passed,
        training will continue from the checkpoint instead of starting from scratch.
        See :paramref:`TrainOutput.params.checkpoint`.
    """

    X_train: List[str]
    y_train: List[str]
    X_valid: List[str]
    y_valid: List[str]
    train_batch_size: int = 32
    valid_batch_size: int = 8
    num_train_epochs: int = 3
    checkpoint: Optional[Path] = None

    def labels(self) -> List[str]:
        """
        Returns:
          The set of unique labels in the data.
          Sort and return a list for consistent ordering, in case that matters.
        """
        return sorted(list(set(self.y_train) | set(self.y_valid)))

    def __post_init__(self):
        for X in (self.X_train, self.X_valid):
            validate_X(X)

        for y in self.y_train, self.y_valid:
            validate_y(y)

        for X, y in ((self.X_train, self.y_train), (self.X_valid, self.y_valid)):
            validate_X_y(X, y)

    def metadata(self) -> Dict[str, Any]:
        return {
            "train_batch_size": self.train_batch_size,
            "valid_batch_size": self.valid_batch_size,
            "num_train_epochs": self.num_train_epochs,
            "len_X_train": len(self.X_train),
            "len_y_train": len(self.y_train),
            "len_X_valid": len(self.X_valid),
            "len_y_valid": len(self.y_valid),
            "checkpoint": self.checkpoint,
        }


@dataclass
class TrainOutput(TaskIO):
    """
    Output from model training.  See :meth:`gobbli.model.mixin.TrainMixin.train`.

    Args:
      valid_loss: Loss on the validation dataset.
      valid_accuracy: Accuracy on the validation dataset.
      train_loss:  Loss on the training dataset.
      labels: List of labels present in the training data.
        Used to initialize the model for prediction.
      checkpoint: Path to the best checkpoint from training.
        This may not be a literal filepath in the case of ex. TensorFlow,
        but it should give the user everything they need to run prediction
        using the results of training.
      _console_output: Raw console output from the container used to train the model.
    """

    valid_loss: float
    valid_accuracy: float
    train_loss: float
    labels: List[str]
    checkpoint: Optional[Path] = None
    _console_output: str = ""

    def metadata(self) -> Dict[str, Any]:
        return {
            "valid_loss": float(self.valid_loss),
            "valid_accuracy": float(self.valid_accuracy),
            "train_loss": float(self.train_loss),
            "labels": self.labels,
            "checkpoint": str(self.checkpoint),
        }


@dataclass
class PredictInput(TaskIO):
    """
    Input for generating predictions using a model.  See :meth:`gobbli.model.mixin.PredictMixin.predict`.

    Args:
      X: Documents to have labels predicted for.
      labels: See :paramref:`TrainOutput.params.labels`.
      predict_batch_size: Number of documents to predict in each batch.
      checkpoint: Checkpoint containing trained weights for the model.
        See :paramref:`TrainOutput.params.checkpoint`.
    """

    X: List[str]
    labels: List[str]
    predict_batch_size: int = 32
    checkpoint: Optional[Path] = None

    def __post_init__(self):
        validate_X(self.X)

    def metadata(self) -> Dict[str, Any]:
        return {
            "predict_batch_size": self.predict_batch_size,
            "labels": self.labels,
            "checkpoint": str(self.checkpoint),
            "len_X": len(self.X),
        }


@dataclass
class PredictOutput(TaskIO):
    """
    Output from generating predictions using a model.  See :meth:`gobbli.model.mixin.PredictMixin.predict`.

    Args:
      y_pred_proba: A dataframe containing the predicted probablity for each class.
        There is a row for each observation and a column for each class.
      _console_output: Raw console output from the container used to generate predictions.
    """

    y_pred_proba: pd.DataFrame
    _console_output: str = ""

    @property
    def y_pred(self) -> List[str]:
        """
        Returns:
          The predicted class for each observation.
        """
        return pred_prob_to_pred_label(self.y_pred_proba)

    def __post_init__(self):
        validate_y(self.y_pred)

    def metadata(self) -> Dict[str, Any]:
        return {"len_y_pred": self.y_pred_proba.shape[0]}


@enum.unique
class EmbedPooling(enum.Enum):
    """
    Enum describing all the different pooling methods that can be used
    when generating embeddings.

    Attributes:
      MEAN: Take the mean across all tokens as the embedding for the document.
      NONE: Return the token-wise embeddings for each document.
    """

    MEAN = "mean"
    NONE = "none"


@dataclass
class EmbedInput(TaskIO):
    """
    Input for generating embeddings using a model.  See :meth:`gobbli.model.mixin.EmbedMixin.embed`.

    Args:
      X: Documents to generate embeddings for.
      embed_batch_size: Number of documents to embed at a time.
      pooling: Pooling method to use for resulting embeddings.
      checkpoint: Checkpoint containing trained weights for the model.
        See :paramref:`TrainOutput.params.checkpoint`.
    """

    X: List[str]

    # Number of documents to embed at a time
    embed_batch_size: int = 32

    # Pooling method to use for resulting embeddings
    pooling: EmbedPooling = EmbedPooling.MEAN

    # Checkpoint containing trained weights for the model
    # See TrainOutput.checkpoint
    checkpoint: Optional[Path] = None

    def __post_init__(self):
        validate_X(self.X)

    def metadata(self) -> Dict[str, Any]:
        return {
            "embed_batch_size": self.embed_batch_size,
            "pooling": self.pooling.value,
            "checkpoint": str(self.checkpoint),
            "len_X": len(self.X),
        }


@dataclass
class EmbedOutput(TaskIO):
    """
    Output from generating embeddings.  See :meth:`gobbli.model.mixin.EmbedMixin.embed`.

    Args:
      X_embedded: A list of ndarrays representing the embedding for each document.
        The shape of each array depends on pooling method. ``l`` = length of the document, and
        ``d`` = dimensionality of embedding.

        - Mean pooling (default): ``(d,)``
        - No pooling: ``(l, d)``
      embed_tokens: If pooling strategy is "NONE", this is the list of tokens
        corresponding to each embedding for each document.  Otherwise, it's :obj:`None`.
      _console_output: Raw console output from the container used to generate the embeddings.
    """

    X_embedded: List[np.ndarray]
    embed_tokens: Optional[List[List[str]]] = None
    _console_output: str = ""

    def metadata(self) -> Dict[str, Any]:
        return {"len_X_embedded": len(self.X_embedded)}


def _chunk_tokens(tokens: List[str], window_len: int) -> Iterator[List[str]]:
    for i in range(0, len(tokens), window_len):
        yield tokens[i : i + window_len]


def make_document_windows(
    X: List[str],
    window_len: int,
    y: Optional[List[str]] = None,
    tokenize_method: TokenizeMethod = TokenizeMethod.SPLIT,
    model_path: Optional[Path] = None,
    vocab_size: Optional[int] = None,
) -> Tuple[List[str], List[int], Optional[List[str]]]:
    """
    This is a helper for when you have a dataset with long documents which is going to be
    passed through a model with a fixed max sequence length.  If you don't have enough
    memory to raise the max sequence length, but you don't want to miss out on the information
    in longer documents, you can use this helper to generate a dataset that splits each
    document into windows roughly the size of your ``max_seq_len``.  The resulting dataset can
    then be used to train your model.  You should then use :func:`pool_document_windows` to
    pool the results from downstream tasks (ex. predictions, embeddings).

    Note there may still be some mismatch between the window size and the size as tokenized
    by your model, since some models use custom tokenization methods.

    Args:
      X: List of texts to make windows out of.
      window_len: The maximum length of each window.  This should roughly correspond to
        the ``max_seq_len`` of your model.
      y: Optional list of labels.  If passed, a list of labels for each window (the label
        associated with the window's document) will be returned.
      tokenize_method: :class:`gobbli.util.TokenizeMethod` corresponding to the tokenization
        method to use for determining windows.
      model_path: This argument is used if the tokenization method requires
        training a model; otherwise, it's ignored.  Path for a tokenization model.
        If it doesn't exist, a new tokenization model will be trained and saved at
        the given path.  If it does exist, the existing model will be used.  If no path
        is given, a temporary directory will be created/used and discarded
      vocab_size: Number of terms in the vocabulary for tokenization. May be ignored depending
        on the tokenization method and whether a model is already trained.

    Returns:
      A 3-tuple containing a new list of texts split into windows, a corresponding list
      containing the index of each original document for each window, and (optionally)
      a list containing a label per window.  The index should
      be used to pool the output from the windowed text (see :func:`pool_document_windows`).
    """
    X_windowed = []  # type: List[str]
    X_windowed_indices = []  # type: List[int]
    y_windowed = []  # type: List[str]

    # Create a temp dir in case it's needed
    with tempfile.TemporaryDirectory() as tmpdir:
        tokenize_kwargs = {}  # type: Dict[str, Any]

        if model_path is None:
            model_path = Path(tmpdir) / "tokenizer"

        tokenize_kwargs["model_path"] = model_path

        detokenize_kwargs = tokenize_kwargs.copy()

        if vocab_size is not None:
            tokenize_kwargs["vocab_size"] = vocab_size

        for i, tokens in enumerate(tokenize(tokenize_method, X, **tokenize_kwargs)):
            for window in detokenize(
                tokenize_method, _chunk_tokens(tokens, window_len), **detokenize_kwargs
            ):
                X_windowed.append(window)
                X_windowed_indices.append(i)
                if y is not None:
                    y_windowed.append(y[i])

    if y is not None:
        return X_windowed, X_windowed_indices, y_windowed
    else:
        return X_windowed, X_windowed_indices, None


@enum.unique
class WindowPooling(enum.Enum):
    """
    Enum describing all the different pooling methods that can be used
    when pooling model output from windowed documents.

    Attributes:
      MEAN: Take the mean across all dimensions/classes as the output for the document.
      MAX: Take the max across all dimensions/classes as the output for the document.
      MIN: Take the min across all dimensions/classes as the output for the document.
    """

    MEAN = "mean"
    MAX = "max"
    MIN = "min"


def pool_document_windows(
    unpooled_output: Union[PredictOutput, EmbedOutput],
    window_indices: List[int],
    pooling: WindowPooling = WindowPooling.MEAN,
):
    """
    This helper pools output from a model whose input was document windows generated by
    :func:`make_document_windows`.  The output can be pooled in multiple ways.  See
    :class:`WindowPooling` for more info.

    This function mutates the passed output object to preserve other information in the
    output object.

    Args:
      unpooled_output: The output from the model to be pooled.
      window_indices: A list (size = number of rows in ``unpooled_output``) of integers corresponding
        to the index of the original document for each window.  These are used to group the window
        output appropriately.
      pooling: The method to use for pooling.
    """
    if isinstance(unpooled_output, PredictOutput):
        unpooled_df = unpooled_output.y_pred_proba
    elif isinstance(unpooled_output, EmbedOutput):
        if unpooled_output.embed_tokens is not None:
            raise ValueError(
                "Embedding output must be pooled when pooling document windows."
            )
        unpooled_df = pd.DataFrame(unpooled_output.X_embedded)
    else:
        raise TypeError(
            f"Unsupported type for unpooled_output: '{type(unpooled_output)}'"
        )

    if not unpooled_df.shape[0] == len(window_indices):
        raise ValueError(f"Unpooled output and window indices must have same length")

    unpooled_df.index = window_indices
    unpooled_grp = unpooled_df.groupby(unpooled_df.index)

    if pooling == WindowPooling.MEAN:
        pooled_df = unpooled_grp.mean()
    elif pooling == WindowPooling.MAX:
        pooled_df = unpooled_grp.max()
    elif pooling == WindowPooling.MIN:
        pooled_df = unpooled_grp.min()
    else:
        raise ValueError(f"Unsupported pooling value: {pooling}")

    if isinstance(unpooled_output, PredictOutput):
        unpooled_output.y_pred_proba = pooled_df
    elif isinstance(unpooled_output, EmbedOutput):
        unpooled_output.X_embedded = [arr for arr in pooled_df.values]



import json
import shutil
import tempfile
from pathlib import Path
from typing import Any, Dict

import numpy as np

import gobbli.io
from gobbli.docker import run_container
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import EmbedMixin
from gobbli.util import assert_in, download_archive, escape_line_delimited_texts


def _read_embeddings(output_file: Path) -> np.ndarray:
    embeddings = []
    with open(output_file, "r") as f:
        for line in f:
            embeddings.append(json.loads(line))
    return np.vstack(embeddings)


USE_MODEL_ARCHIVES = {
    "universal-sentence-encoder": "https://tfhub.dev/google/universal-sentence-encoder/4?tf-hub-format=compressed",
    "universal-sentence-encoder-large": "https://tfhub.dev/google/universal-sentence-encoder-large/5?tf-hub-format=compressed",
    "universal-sentence-encoder-multilingual": "https://tfhub.dev/google/universal-sentence-encoder-multilingual/3?tf-hub-format=compressed",
    "universal-sentence-encoder-multilingual-large": "https://tfhub.dev/google/universal-sentence-encoder-multilingual-large/3?tf-hub-format=compressed",
}
"""
A mapping from model names to TFHub URLs.
"universal-sentence-encoder" is a safe default for most situations.
Larger models require more time and GPU memory to run.
"""


class USE(BaseModel, EmbedMixin):
    """
    Wrapper for Universal Sentence Encoder embeddings:
    https://tfhub.dev/google/universal-sentence-encoder/4
    """

    _BUILD_PATH = Path(__file__).parent

    _INPUT_FILE = "input.txt"
    _OUTPUT_FILE = "output.jsonl"

    def init(self, params: Dict[str, Any]):
        """
        See :meth:`gobbli.model.base.BaseModel.init`.

        USE parameters:

        - ``use_model`` (:obj:`str`): Name of a USE model to use.
          See :obj:`USE_MODEL_ARCHIVES` for a listing of available USE models.
        """
        self.use_model = "universal-sentence-encoder"

        for name, value in params.items():
            if name == "use_model":
                assert_in(name, value, set(USE_MODEL_ARCHIVES.keys()))
                self.use_model = value
            else:
                raise ValueError(f"Unknown param '{name}'")

    @property
    def image_tag(self) -> str:
        """
        Returns:
          The Docker image tag to be used for the USE container.
        """
        device = "gpu" if self.use_gpu else "cpu"
        return f"gobbli-use-embeddings-{device}"

    @property
    def weights_dir(self) -> Path:
        """
        Returns:
          Directory containing pretrained weights for this instance.
        """
        return self.class_weights_dir / self.use_model

    def _build(self):
        # Download data if we don't already have it
        if not self.weights_dir.exists():
            with tempfile.TemporaryDirectory() as tmpdir:
                tmp_weights_dir = Path(tmpdir) / self.weights_dir.name
                tmp_weights_dir.mkdir()
                self.logger.info("Downloading pre-trained weights.")
                download_archive(
                    USE_MODEL_ARCHIVES[self.use_model],
                    tmp_weights_dir,
                    filename=f"{self.use_model}.tar.gz",
                )
                shutil.move(tmp_weights_dir, self.weights_dir)
                self.logger.info("Weights downloaded.")

        # Build the docker image
        self.docker_client.images.build(
            path=str(USE._BUILD_PATH),
            tag=self.image_tag,
            **self._base_docker_build_kwargs,
        )

    def _embed(
        self, embed_input: gobbli.io.EmbedInput, context: ContainerTaskContext
    ) -> gobbli.io.EmbedOutput:

        if embed_input.pooling == gobbli.io.EmbedPooling.NONE:
            raise ValueError(
                "Universal Sentence Encoder does sentence encoding, so pooling is required."
            )

        (context.host_input_dir / USE._INPUT_FILE).write_text(
            escape_line_delimited_texts(embed_input.X)
        )

        cmd = (
            "python use.py"
            f" --input-file={context.container_input_dir / USE._INPUT_FILE}"
            f" --output-file={context.container_output_dir / USE._OUTPUT_FILE}"
            f" --module-dir={BaseModel._CONTAINER_WEIGHTS_PATH}"
            f" --batch-size={embed_input.embed_batch_size}"
        )

        container_logs = run_container(
            self.docker_client,
            self.image_tag,
            cmd,
            self.logger,
            **self._base_docker_run_kwargs(context),
        )

        return gobbli.io.EmbedOutput(
            X_embedded=_read_embeddings(context.host_output_dir / USE._OUTPUT_FILE),
            _console_output=container_logs,
        )



import pytest

from gobbli.augment.bert import BERTMaskedLM
from gobbli.test.util import model_test_dir


@pytest.mark.parametrize(
    "params,exception",
    [
        # Unknown param
        ({"unknown": None}, ValueError),
        # Bad type (diversity)
        ({"diversity": 2}, TypeError),
        # Bad type (batch size)
        ({"batch_size": 2.5}, TypeError),
        # Bad type (n_probable)
        ({"n_probable": 2.5}, TypeError),
        # Bad value (diversity)
        ({"diversity": 0.0}, ValueError),
        # Bad value (batch_size)
        ({"batch_size": 0}, ValueError),
        # Bad value (n_probable)
        ({"n_probable": 0}, ValueError),
        # OK values
        ({"diversity": 0.5, "n_probable": 3, "batch_size": 16}, None),
    ],
)
def test_init(params, exception):
    if exception is None:
        BERTMaskedLM(**params)
    else:
        with pytest.raises(exception):
            BERTMaskedLM(**params)


def test_bertmaskedlm_augment(model_gpu_config, gobbli_dir):
    model = BERTMaskedLM(
        data_dir=model_test_dir(BERTMaskedLM), load_existing=True, **model_gpu_config
    )
    model.build()

    times = 5
    new_texts = model.augment(["This is a test."], times=times)
    assert len(new_texts) == times



# coding=utf-8
# Copyright (c) Microsoft. All rights reserved.
import csv
import json
import logging
import os
import pickle as pkl
import random
import string
import sys
from shutil import copyfile

import numpy as np

import torch
from pytorch_pretrained_bert.tokenization import BertTokenizer

UNK_ID=100
BOS_ID=101


def tokenize_sample(bert_tokenizer, sample, max_seq_len, has_labels, labels):
    """
    Run tokenization for a singel-sentence task.
    """
    X = bert_tokenizer.tokenize(sample['X'])
    if has_labels:
        y = sample['y']
    if len(X) >  max_seq_len - 3:
        X = X[:max_seq_len - 3]
    input_ids = bert_tokenizer.convert_tokens_to_ids(['[CLS]'] + X + ['[SEP]'])
    token_type_ids = [0] * ( len(X) + 2)
    tokenized = {'input_ids': input_ids, 'token_type_ids': token_type_ids}
    if has_labels:
        tokenized['y'] = labels[y]
    return tokenized


class GobbliBatchGen:
    def __init__(self, path, batch_size=32, gpu=True, labels=None,
                 has_labels=True, is_train=True, dropout_w=0.005, maxlen=128):
        self.batch_size = batch_size
        self.has_labels = has_labels
        self.gpu = gpu
        self.labels = labels
        self.is_train = is_train
        # Explicit cache dir required for some reason -- default doesn't exist in the docker
        # container, maybe?
        self.bert_tokenizer = BertTokenizer.from_pretrained('bert-base-uncased', cache_dir='/tmp')
        self.data = self.load(path, maxlen, has_labels)
        if self.is_train:
            indices = list(range(len(self.data)))
            random.shuffle(indices)
            data = [self.data[i] for i in indices]
        self.data = GobbliBatchGen.make_batches(self.data, batch_size)
        self.offset = 0
        self.dropout_w = dropout_w

    @staticmethod
    def make_batches(data, batch_size=32):
        return [data[i:i + batch_size] for i in range(0, len(data), batch_size)]

    def load(self, path, maxlen, has_labels):
        data = []
        with open(path, 'r', encoding='utf-8') as f:
            reader = csv.DictReader(f)
            for row in reader:
                sample = tokenize_sample(self.bert_tokenizer, row, maxlen, has_labels, self.labels)
                if self.is_train:
                    # TODO why is this needed?
                    if len(sample['input_ids']) > maxlen:
                        continue
                data.append(sample)

        print('Loaded {} samples'.format(len(data)))
        return data

    def reset(self):
        if self.is_train:
            indices = list(range(len(self.data)))
            random.shuffle(indices)
            self.data = [self.data[i] for i in indices]
        self.offset = 0

    def __random_select__(self, arr):
        if self.dropout_w > 0:
            return [UNK_ID if random.uniform(0, 1) < self.dropout_w else e for e in arr]
        else: return arr

    def __len__(self):
        return len(self.data)

    def patch(self, v):
        v = v.cuda(async=True)
        return v

    @staticmethod
    def todevice(v, device):
        v = v.to(device)
        return v


    def __iter__(self):
        while self.offset < len(self):
            batch = self.data[self.offset]
            batch_size = len(batch)
            tok_len = max(len(x['input_ids']) for x in batch)
            input_ids = torch.LongTensor(batch_size, tok_len).fill_(0)
            token_type_ids = torch.LongTensor(batch_size, tok_len).fill_(0)
            attention_mask = torch.LongTensor(batch_size, tok_len).fill_(0)

            for i, sample in enumerate(batch):
                select_len = min(len(sample['input_ids']), tok_len)
                tok = sample['input_ids']
                if self.is_train:
                    tok = self.__random_select__(tok)
                input_ids[i, :select_len] = torch.LongTensor(tok[:select_len])
                token_type_ids[i, :select_len] = torch.LongTensor(sample['token_type_ids'][:select_len])
                attention_mask[i, :select_len] = torch.LongTensor([1] * select_len)

            batch_data = [input_ids, token_type_ids, attention_mask]

            if self.has_labels:
                labels = [sample['y'] for sample in batch]
                batch_data.append(torch.LongTensor(labels))

            if self.gpu:
                for i, item in enumerate(batch_data):
                    batch_data[i] = self.patch(item.pin_memory())

            self.offset += 1

            if self.has_labels:
                yield batch_data[0], batch_data[1], batch_data[2], batch_data[3]
            else:
                yield batch_data[0], batch_data[1], batch_data[2]



import pandas as pd

from gobbli.inspect.evaluate import ClassificationError, ClassificationEvaluation


def test_classification_evaluation():
    results = ClassificationEvaluation(
        labels=["a", "b"],
        X=["a1", "a2", "a3", "a4", "b1", "b2", "b3", "b4", "b5"],
        y_true=["a", "a", "a", "a", "b", "b", "b", "b", "b"],
        y_pred_proba=pd.DataFrame(
            {
                "a": [0.7, 0.3, 0.4, 0.6, 0.9, 0.3, 0.6, 0.4, 0.7],
                "b": [0.3, 0.7, 0.6, 0.4, 0.1, 0.7, 0.4, 0.6, 0.3],
            }
        ),
    )

    # Ensure predicted labels are calculated correctly
    expected_y_pred = ["a", "b", "b", "a", "a", "b", "a", "b", "a"]
    assert results.y_pred == expected_y_pred

    # Ensure errors are calculated correctly
    # Error observations are a2, a3, b1, b3, b5
    ea2 = ClassificationError(X="a2", y_true="a", y_pred_proba={"a": 0.3, "b": 0.7})
    ea3 = ClassificationError(X="a3", y_true="a", y_pred_proba={"a": 0.4, "b": 0.6})
    eb1 = ClassificationError(X="b1", y_true="b", y_pred_proba={"a": 0.9, "b": 0.1})
    eb3 = ClassificationError(  # noqa: F841
        X="b3", y_true="b", y_pred_proba={"a": 0.6, "b": 0.4}
    )
    eb5 = ClassificationError(X="b5", y_true="b", y_pred_proba={"a": 0.7, "b": 0.3})

    # Cut off at k=2 and ensure the correct errors are in there and ordered correctly
    errors = results.errors(k=2)

    # Expected errors
    a_false_positives = [eb1, eb5]
    a_false_negatives = [ea2, ea3]
    b_false_positives = a_false_negatives
    b_false_negatives = a_false_positives

    a_errors = errors["a"]
    b_errors = errors["b"]
    assert a_errors[0] == a_false_positives
    assert a_errors[1] == a_false_negatives
    assert b_errors[0] == b_false_positives
    assert b_errors[1] == b_false_negatives



from typing import Any, Dict

import numpy as np
import pandas as pd

import gobbli.io
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import PredictMixin, TrainMixin


class MajorityClassifier(BaseModel, TrainMixin, PredictMixin):
    """
    Simple classifier that returns the majority class from the training set.

    Useful for ensuring user code works with the gobbli input/output format
    without having to build a time-consuming model.
    """

    def init(self, params: Dict[str, Any]):
        self.majority_class: Any = None

    def _build(self):
        """
        No build step required for this model.
        """

    def _train(
        self, train_input: gobbli.io.TrainInput, context: ContainerTaskContext
    ) -> gobbli.io.TrainOutput:
        """
        Determine the majority class.
        """
        unique_values, value_counts = np.unique(train_input.y_train, return_counts=True)
        self.majority_class = unique_values[value_counts.argmax(axis=0)]

        y_train_pred = np.full_like(train_input.y_train, self.majority_class)
        train_loss = np.sum(y_train_pred != train_input.y_train)

        y_valid_pred = np.full_like(train_input.y_valid, self.majority_class)
        valid_loss = np.sum(y_valid_pred != train_input.y_valid)
        valid_accuracy = valid_loss / y_valid_pred.shape[0]

        return gobbli.io.TrainOutput(
            valid_loss=valid_loss,
            valid_accuracy=valid_accuracy,
            train_loss=train_loss,
            labels=train_input.labels(),
        )

    def _predict(
        self, predict_input: gobbli.io.PredictInput, context: ContainerTaskContext
    ) -> gobbli.io.PredictOutput:
        """
        Predict based on our learned majority class.
        """
        pred_proba_df = pd.DataFrame(
            {
                label: 1 if label == self.majority_class else 0
                for label in predict_input.labels
            },
            index=range(len(predict_input.X)),
        )

        return gobbli.io.PredictOutput(y_pred_proba=pred_proba_df)



import logging
import os
from abc import ABC
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import ray

from gobbli.dataset.base import BaseDataset
from gobbli.model.sklearn import SKLearnClassifier
from gobbli.util import generate_uuid, gobbli_dir, is_dir_empty, write_metadata


def experiment_dir() -> Path:
    return gobbli_dir() / "experiment"


def init_worker_env(
    gobbli_dir: Optional[Path] = None, log_level: Union[int, str] = logging.WARNING
) -> logging.Logger:
    """
    Initialize environment on a ray worker.

    Args:
        gobbli_dir: Used as the value of the GOBBLI_DIR environment variable;
        determines where gobbli data is stored on the worker's filesystem.
        log_level: Level for logging coming from the worker.
    """
    if gobbli_dir is not None:
        os.environ["GOBBLI_DIR"] = str(gobbli_dir)

    logging.basicConfig(level=log_level)

    # Provide a logger for our workers
    # Workers should only log using loggers they've created to avoid
    # logger pickling, which generally doesn't work
    # https://stackoverflow.com/questions/55272066/how-can-i-use-the-python-logging-in-ray
    return logging.getLogger(__name__)


def init_gpu_config() -> Tuple[bool, str]:
    """
    Determine the GPU configuration from the current ray environment
    on a worker.

    Returns:
      2-tuple: whether GPU should be used and a comma-separated string
      containing the ids of the GPUs that should be used
    """
    try:
        gpu_ids = ray.get_gpu_ids()
    except Exception as e:
        # This message is either 'ray.get_gpu_ids() currently does not work in PYTHON MODE'
        # or '... in LOCAL MODE' depending on the version of ray installed
        if "ray.get_gpu_ids() currently does not work in" in str(e):
            gpu_ids = []
        else:
            raise
    use_gpu = len(gpu_ids) > 0
    nvidia_visible_devices = ",".join(str(i) for i in gpu_ids)
    return use_gpu, nvidia_visible_devices


def get_worker_ip() -> str:
    """
    Determine the IP address of the current ray worker.
    Returns:
      A string containing the IP address.
    """
    global_worker = ray.worker.get_global_worker()
    return getattr(global_worker, "node_ip_address", "127.0.0.1")


def is_ray_local_mode() -> bool:
    """
    Determine whether the current global ray worker is running in local mode.
    Returns:
      True if we're running in local mode.
    """
    return ray.worker.get_global_worker().mode == ray.worker.LOCAL_MODE


class BaseExperiment(ABC):
    """
    Base class for all derived Experiments.
    """

    _METADATA_FILENAME = "gobbli-experiment-meta.json"

    def __init__(
        self,
        model_cls: Any,
        dataset: Union[Tuple[List[str], List[str]], BaseDataset],
        test_dataset: Optional[Tuple[List[str], List[str]]] = None,
        data_dir: Optional[Path] = None,
        name: Optional[str] = None,
        param_grid: Optional[Dict[str, List[Any]]] = None,
        task_num_cpus: int = 1,
        task_num_gpus: int = 0,
        worker_gobbli_dir: Optional[Path] = None,
        worker_log_level: Union[int, str] = logging.WARNING,
        limit: Optional[int] = None,
        overwrite_existing: bool = False,
        ignore_ray_initialized_error: bool = False,
        distributed: bool = False,
        ray_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """
        Construct an experiment.

        Args:
          model_cls: The class of model to be used for the experiment.
          dataset: Dataset to be used for the experiment.  Can be either a 2-tuple
            containing a list of texts and a corresponding list of labels
            or a :class:`gobbli.dataset.base.BaseDataset`.
          test_dataset: An optional separate dataset to be used for calculating test metrics.
            If passed, should be a 2-tuple containing a list of texts and corresponding list
            of labels.  If not passed, a test dataset will be automatically split out of
            the `dataset`.
          data_dir: Optional path to a directory used to store data for the experiment.
            If not given, a directory under GOBBLI_DIR will be created and used.
          name: A descriptive name for the experiment, used to label directories
            in the filesystem.  If not passed, a random name will be generated and used.
            The name must be unique (i.e., there should not be another experiment with the
            same name).
          param_grid: Optional grid of parameters.  If passed, it should be a dictionary
            with keys being valid parameter names for the passed model and values being lists
            of parameter values.  Every combination of parameter values will be tried in the
            experiment, and the results for the best combination will be returned.  If not passed,
            only the model's default parameters will be used.
          task_num_cpus: Number of CPUs to reserve per task.
          task_num_gpus: Number of GPUs to reserve per task.
          worker_gobbli_dir: Directory to use for gobbli file storage by workers.
          worker_log_level: Logging level to use for logs output by workers running
            training tasks.
          limit: Read up to this many rows from the passed dataset.  Useful for debugging.
          overwrite_existing: If True, don't fail if there's an existing experiment in
            the same directory.
          ignore_ray_initialized_error: If True, don't error when a ray connection is already
            initialized; instead, shut it down and restart it with the passed `ray_kwargs`.
          distributed: If True, run the ray cluster assuming workers are distributed over
            multiple nodes.  This requires model weights for all trials to fit in the ray
            object store, which requires a lot of memory.  If False, run the ray cluster
            assuming all workers are on the master node, and weights will be passed around as
            filepaths; an error will be thrown if a remote worker tries to run a task.
          ray_kwargs: Dictionary containing keyword arguments to be passed directly to
            :func:`ray.init`.  By default, a new ray cluster will be initialized on the current
            node using all available CPUs and no GPUs, but these arguments can be used to connect
            to a remote cluster, limit resource usage, and much more.
        """
        self.model_cls = model_cls
        self.worker_gobbli_dir = worker_gobbli_dir

        self.name = name
        if self.name is None:
            self.name = generate_uuid()

        if data_dir is None:
            self._data_dir = experiment_dir() / self.__class__.__name__ / self.name
        else:
            self._data_dir = data_dir
        self._data_dir.mkdir(parents=True, exist_ok=True)

        if not overwrite_existing and not is_dir_empty(self._data_dir):
            raise ValueError(f"Experiment already exists for name '{self.name}'")

        if isinstance(dataset, BaseDataset):
            self.X = dataset.X_train() + dataset.X_test()
            self.y = dataset.y_train() + dataset.y_test()
        elif isinstance(dataset, tuple):
            if len(dataset) != 2:
                raise ValueError(
                    f"`dataset` must be a 2-tuple, got length {len(dataset)}"
                )
            self.X, self.y = dataset
        else:
            raise TypeError(f"Invalid type for dataset: {type(dataset)}")

        self.X_test = None  # type: Optional[List[str]]
        self.y_test = None  # type: Optional[List[str]]
        if test_dataset is not None:
            if not (isinstance(dataset, tuple) and len(dataset) == 2):
                raise ValueError(f"`test_dataset` must be a 2-tuple")
            self.X_test, self.y_test = test_dataset

        if limit is not None:
            self.X = self.X[:limit]
            self.y = self.y[:limit]

        self.param_grid = param_grid
        if param_grid is None:
            self.param_grid = {}

        self.task_num_cpus = task_num_cpus
        self.task_num_gpus = task_num_gpus
        self.worker_log_level = worker_log_level
        self.distributed = distributed

        if self.model_cls is SKLearnClassifier and distributed:
            raise ValueError(
                "The scikit-learn classifier is not supported for distributed "
                "experiments, since it needs to load a pickle from a file path "
                "which may not be on a given worker node."
            )

        _ray_kwargs = ray_kwargs
        if _ray_kwargs is None:
            _ray_kwargs = {}

        # We may have an existing ray connection active -- throw an error or
        # clear it out to ensure it's re-initialized with the passed params
        if ray.is_initialized():
            if ignore_ray_initialized_error:
                ray.shutdown()
            else:
                raise RuntimeError(
                    "A ray connection is already initialized. To ignore this error"
                    " and shut down the existing connection, pass"
                    " `ignore_ray_initialized_error=True`."
                )

        ray.init(**_ray_kwargs)

        metadata = {
            "model": model_cls.__name__,
            "len_X": len(self.X),
            "len_y": len(self.y),
            "param_grid": self.param_grid,
        }

        write_metadata(metadata, self.metadata_path)

    @property
    def metadata_path(self) -> Path:
        """
        Returns:
          The path to the experiment's metadata file containing information about the
          experiment parameters.
        """
        return self.data_dir() / BaseExperiment._METADATA_FILENAME

    def data_dir(self) -> Path:
        """
        Returns:
          The main data directory unique to this experiment.
        """
        return self._data_dir



# Copyright (c) Microsoft. All rights reserved.

from .vocab import Vocabulary
from .metrics import compute_acc, compute_f1, compute_mcc, compute_pearson, compute_spearman, compute_cross_entropy
# scitail
ScitailLabelMapper = Vocabulary(True)
ScitailLabelMapper.add('neutral')
ScitailLabelMapper.add('entails')

# label map
SNLI_LabelMapper = Vocabulary(True)
SNLI_LabelMapper.add('contradiction')
SNLI_LabelMapper.add('neutral')
SNLI_LabelMapper.add('entailment')

# qnli
QNLILabelMapper = Vocabulary(True)
QNLILabelMapper.add('not_entailment')
QNLILabelMapper.add('entailment')

GLOBAL_MAP = {
 'scitail': ScitailLabelMapper,
 'mnli': SNLI_LabelMapper,
 'snli': SNLI_LabelMapper,
 'qnli': QNLILabelMapper,
 'qnnli': QNLILabelMapper,
 'rte': QNLILabelMapper,
 'diag': SNLI_LabelMapper,
}

# number of class
DATA_META = {
 'mnli': 3,
 'snli': 3,
 'scitail': 2,
 'qqp': 2,
 'qnli': 2,
 'qnnli': 1,
 'wnli': 2,
 'rte': 2,
 'mrpc': 2,
 'diag': 3,
 'sst': 2,
 'stsb': 1,
 'cola': 2,
}

DATA_TYPE = {
 'mnli': 0,
 'snli': 0,
 'scitail': 0,
 'qqp': 0,
 'qnli': 0,
 'qnnli': 0,
 'wnli': 0,
 'rte': 0,
 'mrpc': 0,
 'diag': 0,
 'sst': 1,
 'stsb': 0,
 'cola': 1,
}

DATA_SWAP = {
 'mnli': 0,
 'snli': 0,
 'scitail': 0,
 'qqp': 1,
 'qnli': 0,
 'qnnli': 0,
 'wnli': 0,
 'rte': 0,
 'mrpc': 0,
 'diag': 0,
 'sst': 0,
 'stsb': 0,
 'cola': 0,
}

# classification/regression
TASK_TYPE = {
 'mnli': 0,
 'snli': 0,
 'scitail': 0,
 'qqp': 0,
 'qnli': 0,
 'qnnli': 0,
 'wnli': 0,
 'rte': 0,
 'mrpc': 0,
 'diag': 0,
 'sst': 0,
 'stsb':1,
 'cola': 0,
}

METRIC_META = {
 'mnli': [0],
 'snli': [0],
 'scitail': [0],
 'qqp': [0, 1],
 'qnli':[0],
 'qnnli': [0],
 'wnli': [0],
 'rte': [0],
 'mrpc': [0, 1],
 'diag': [0],
 'sst': [0],
 'stsb': [3, 4],
 'cola': [0, 2],
}

METRIC_NAME = {
 0: 'ACC',
 1: 'F1',
 2: 'MCC',
 3: 'Pearson',
 4: 'Spearman',
 5: 'Loss',
}

METRIC_FUNC = {
 0: compute_acc,
 1: compute_f1,
 2: compute_mcc,
 3: compute_pearson,
 4: compute_spearman,
    5: compute_cross_entropy,
}

SAN_META = {
    'mnli': 1,
    'snli': 1,
    'scitail': 1,
    'qqp': 1,
    'qnli': 1,
    'qnnli': 1,
    'wnli': 1,
    'rte': 1,
    'mrpc': 1,
    'diag': 0,
    'sst': 0,
    'stsb': 0,
    'cola': 0,
}

def generate_decoder_opt(task, max_opt):
    assert task in SAN_META
    opt_v = 0
    if SAN_META[task] and max_opt < 3:
        opt_v = max_opt
    return opt_v
